argStr n tn = "arg" ++ show n ++ "t" ++ show tn
letStr n tn = "larg" ++ show n ++ "t" ++ show tn
newLambdaArg argT names argUsedness from = do
   context <- gets pdContext
   let depth = 1 + length context
   let impls = zipWith3 (\t name tn -> Impl {
      implType = t,
      implCode = hsAtom $ argStr depth tn,
      implDeps = Set.singleton depth,
      implName = name,
      implUsed = argUsedness
      } ) argT (transposeMaybe names) [1..]
   let newArg = Args impls LambdaArg from
   modify $ \s -> s { pdContext=newArg:context }
   return newArg
transposeMaybe names = maybe (repeat Nothing) (map Just) names
newLetArg names argUsedness (Impl _ defHs defDepth _ usedness) defTypes from = do
   context <- gets pdContext
   nextLetId <- gets pdNextUniqLetId
   modify $ \s -> s { pdNextUniqLetId=nextLetId+1 }
   let impls = zipWith3 (\t name tn -> Impl t (hsAtom $ letStr nextLetId tn) defDepth name argUsedness) defTypes (transposeMaybe names) [1..]
   let newArg = Args impls (LetArg defHs) from
   modify $ \s -> s { pdContext=newArg:context }
   return newArg
argn deBruijnIndex = do
   context <- gets pdContext
   flattenedArgs <- getAllArgs
   case at flattenedArgs (deBruijnIndex-1) of
      Nothing -> parseError $ "Attempt to access " ++ (snd $ indexToOp (deBruijnIndex-1)) ++ debugContext context
      Just impl -> setUsed impl
getArgByName = do
   allArgs <- getAllArgs
   getArgByNameH 0 allArgs
   where
      getArgByNameH _ [] = return Nothing
      getArgByNameH n (impl:rest) =
         case implName impl of
            Just argName -> do
               code <- gets pdCode
               case onlyCheckMatchIdentifier code of
                  Just (name, nextCode) | name == argName -> do
                     modify $ \s -> s { pdCode = nextCode }
                     appendRep $ indexToOp n
                     argn (n+1) >>= return.Just
                  otherwise -> getArgByNameH (n+1) rest
            otherwise -> getArgByNameH (n+1) rest
setUsed impl = do
   context <- gets pdContext
   let newContext = map (\arg -> arg { argsImpls=setUsedImpl (argsImpls arg) } ) context
   modify $ \s -> s { pdContext=newContext }
   return impl where
   setUsedImpl = map (\argImpl ->
      if implCode argImpl == implCode impl then
         argImpl { implUsed=UsedArg }
      else argImpl)
argImplicit = do
   if ?isSimple
   then parseError "Expecting more expressions at EOF"
   else do
      allArgs <- getAllArgs
      modify $ \s -> s { pdImplicitArgUsed = True }
      argn $ 1 + (fromMaybe 0 $ findIndex ((==UnusedArg).implUsed) allArgs)
flattenArg (Args impls (LambdaArg) _) = impls
flattenArg (Args impls (LetArg _) _) = tail impls
getAllArgs = do
   context <- gets pdContext
   return $ concatMap flattenArg context
addLambda arg (Impl t body d _ _) = Impl {
   implType = t,
   implCode = hsFn (map implCode $ argsImpls arg) body,
   implDeps = Set.difference d (getArgDeps arg),
   implName = Nothing,
   implUsed = UsednessDoesntMatter }
popArg depth impl = do
   context <- gets pdContext
   let (finalImpl, finalContext) = mapAccumL maybePopIt impl context
   modify $ \s -> s { pdContext=concat finalContext }
   return finalImpl where
   maybePopIt impl eachArg
      | Set.member depth (getArgDeps eachArg) = (popIt eachArg impl, [])
      | otherwise = (impl, [eachArg])
   popIt (Args _ LambdaArg _) impl = impl
   popIt (Args varImpls (LetArg refHs) _) (Impl retT bodyHs dep _ _) =
      Impl retT (hsLet (map implCode varImpls) refHs bodyHs) dep Nothing UsednessDoesntMatter
getArgDeps (Args [] _ _) = implDeps noArgsUsed
debugContext context = "\nContext:\n" ++ (unlines $ snd $ mapAccumL (\count arg ->
   let args = flattenArg arg in
   (count+length args, unlines $ showArgType arg : zipWith (\n a->("  "++showArg n a)) [count..] args)
   ) 0 $ filter (not.null.flattenArg) context)
showArgType (Args _ LambdaArg from) = "LambdaArg " ++ from
showArgType (Args _ (LetArg _) from) = "LetArg " ++ from
explainUsedness UnusedArg = "(unused so far, prioritized for implicit args)"
explainUsedness UsednessDoesntMatter = "usedness doesn't matter, should be impossible"
indexToOp = fromMaybe (error "deBruijn index too high") . (at deBruijnArgReps) where
deBruijnArgReps = [(replicate unary 6 ++ [nib], replicate unary ';' ++ sym)
padToEvenNibbles s = s ++ replicate (length s `mod` 2) uselessOp
compile finishFn separator cArgs input = evalState doCompile $ blankRep (consumeWhitespace input) args 0 where
   args =
      [ Args (Impl undefined (hsAtom"_") (Set.singleton 0) Nothing UsedArg:letArgs)
         (LetArg $ hsAtom $ "(undefined," ++ intercalate "," letDefs ++ ")") "inputs"]
   mainLets = cArgs ++
      [ ("fstInt", VInt, "if datOverride then dat else fromMaybe 100 $ at intList 0")
      , ("fstLine", vstr, "fromMaybe printables $ at strLines 0")
      , ("ints", VList [VInt], "intList")
      , ("sndInt", VInt, "fromMaybe 1000 $ at intList 1")
      , ("allLines", VList [vstr], "strLines")
      , ("allInput", vstr, "input")
      , ("intMatrix", VList [VList [VInt]], "intMatrix_")
      , ("sndLine", vstr, "fromMaybe [] $ at strLines 1")
      ]
   letArgs = zipWith (\(name, vt, _) ind -> noArgsUsed {
      implType=vt, implCode=hsAtom name, implName=Just name, implUsed=if ind>length cArgs then OptionalArg else UnusedArg
      }) mainLets [1..]
   letDefs = map (\(_, _, hsDef) -> hsDef) mainLets
   doCompile = do
      impl1 <- get1Value
      (dat,body) <- if separator == "" then mainCombiner (impl1,False)
         else testCombiner $ fst $ finishIt impl1
      context <- gets pdContext
      impl <- popArg 0 body
      nib <- gets getNib
      lit <- gets getLit
      dataUsed <- gets pdDataUsed
      let [fstIntUsed,fstLineUsed,intsUsed,sndIntUsed,allInputUsedAsLines,allInputUsed,allInputUsedAsInts,sndLineUsed] =
            drop (length cArgs) $ getInputsUsedness context
      let useDataInsteadOfFirstIntInput = isJust dat && not dataUsed
      let autoMapCode = (if separator == "," then "in intercalate [newli]" else "in concat $ map finishLn")++" $ flip map autoMapList $ \\"
      let autoMap = if ?isSimple then "" else
            if allInputUsed || allInputUsedAsInts || allInputUsedAsLines then ""
            else if sndLineUsed then "let autoMapList = (listOr [[]] (chunksOf 2 strLines)) "++autoMapCode++"strLines -> "
            else if fstLineUsed then "let autoMapList = (listOr [[]] (chunksOf 1 strLines)) "++autoMapCode++"strLines -> "
            else if intsUsed then "let autoMapList = if length intMatrix_ > 1 && (any ((>1).length) intMatrix_) then intMatrix_ else [intList] "++autoMapCode++"intList -> "
            else if sndIntUsed then "let autoMapList = (listOr [[]] (chunksOf 2 intList)) "++autoMapCode++"intList -> "
            else if fstIntUsed && not useDataInsteadOfFirstIntInput then "let autoMapList = (listOr [[]] (chunksOf 1 intList)) "++autoMapCode++"intList -> "
            else ""
      let finalImpl = app1Hs ("let intMatrix_=filter (not . null) (map (asInts.sToA) (lines $ aToS input));\
            \strLines=map sToA $ lines $ aToS input;\
            \intList=concat intMatrix_;\
            \datOverride="++show useDataInsteadOfFirstIntInput ++ ";\
            \dat="++show (fromMaybe 0 dat) ++";\
            \autoMapList=[];\
            \in "++autoMap++"let (firstSep,secondSep)=if length autoMapList > 1 then ([],[space]) else ([space],[newli]) in") impl
      warnings <- gets pdLitWarnings
      return (finalImpl, nib, lit, warnings)
   finishIt impl = let (finishF,isLastList) = finishFn (implType impl) in
      ((app1Hs finishF impl) { implType = vstr }, isLastList)
   join2 extraSep impl1 impl2 = let
         (finished, isList) = finishIt impl2
         sep = if extraSep && separator == "" then "secondSep" else "sToA" ++show separator
      in (applyImpl (app1Hs ("(\\a b->a++"++sep++"++b)") impl1) finished, isList)
   mainCombiner (prev,isLastList) = do
      code <- gets pdCode
      let (finishedPrev,isPrevList) = finishIt prev
      if empty code then return (Nothing, finishedPrev)
      else do
         context <- gets pdContext
         dataUsed <- gets pdDataUsed
         doData <- if dataUsed then
            return True
            then match tildaOp
         else return False
         if doData then do
            dat <- parseDataExpr
            return (Just dat,finishedPrev)
         else do
            case implType prev of
               VInt | not ?isSimple -> do
                  (impl1,argsUsed) <- getLambdaValue 1 [VInt,VInt] OptionalArg "implicit op int" False
                  if last argsUsed then do
                     mainCombiner ((applyImpl (app1Hs ("("++foldr1Fn "(foldl1.flip)" "id" [VList [VInt], implType impl1]++")") (app1Hs "(\\x->[1..x])" prev)) impl1) { implType = VInt }, False)
                  else if argsUsed == [True,False] then do
                     mainCombiner ((applyImpl (app1Hs "(\\a f -> map (\\y->f (y,())) [1..a])" prev) impl1) { implType = VList $ ret $ implType impl1 }, False)
                  else do
                     rhsImpl <- convertPairToLet UnusedArg (app1Hs (fillAccums 2 0) impl1) (ret $ implType impl1) "tuple in possible implicit op"
                     afterCode <- gets pdCode
                     mainCombiner $ join2 extraSep finishedPrev rhsImpl
               VList e | not ?isSimple && e /= [VChr] -> do
                  (impl1,argsUsed) <- getLambdaValue 1 (e++e) OptionalArg "implicit op list" False
                  if or $ drop (length e) argsUsed then do
                     impl <- convertPairToLet UnusedArg (applyImpl (applyImpl (noArgsUsed { implCode=hsParen $ hsAtom $ foldr1Fn "(foldl1.flip)" "id" [implType prev, implType impl1] }) prev) impl1) e "tuple in implicit op foldl"
                     mainCombiner (impl,False)
                  else if or $ take (length e) argsUsed then do
                     mainCombiner $ ((applyImpl (applyImpl (noArgsUsed { implCode=hsParen $ hsAtom $ mapFn [implType prev, implType impl1] }) prev) (app1Hs (fillAccums (length e) (2*length e)) impl1)) { implType = VList $ ret $ implType impl1 }, False)
                  else do
                     rhsImpl <- convertPairToLet UnusedArg (app1Hs (fillAccums (2*length e) (2*length e)) impl1) (ret $ implType impl1) "tuple in possible implicit op"
                     afterCode <- gets pdCode
                     mainCombiner $ join2 extraSep finishedPrev rhsImpl
               otherwise -> do
                  impl1 <- get1Value
                  afterCode <- gets pdCode
                  mainCombiner $ join2 extraSep finishedPrev impl1
   testCombiner prev = do
      code <- gets pdCode
      if empty code then return (Nothing, prev)
      else do
         impl1 <- get1Value
            (Impl _ _ _ _ _) -> testCombiner $ fst (join2 False prev impl1)
   getInputsUsedness context = tail $ map ((==UsedArg).implUsed) $ argsImpls $ last context
applyImpl impl1 impl2 | elem (implType impl2) [OptionYes, OptionNo] = impl1
applyImpl (Impl t1 hs1 d1 _ _) (Impl t2 hs2 d2 _ _) =
   let hs2annotated = case toHsType t2 of
         Nothing -> hsParen hs2 in
   Impl t1 (hsApp hs1 hs2annotated) (Set.union d1 d2) undefined undefined
makePairs fromTypes args = foldl applyImpl initImpl args where
   toTypes = map implType args
   initImpl = noArgsUsed { implType=VFn fromTypes toTypes, implCode=hsAtom pairMakerHs }
   pairMakerHs = if length args == 1 then "" else "("++replicate (length $ tail args) ','++")"
putAddRep (ParseData code context nib lit dataUsed implicitArgUsed warnings letid) = do
   appendRepH (nib,lit)
   dataUsed1 <- gets pdDataUsed
   context1 <- gets pdContext
   warnings1 <- gets pdLitWarnings
   modify $ \s -> s { pdCode=code, pdContext=unionUsed context1 context, pdDataUsed=dataUsed1 || dataUsed, pdLitWarnings = warnings ++ warnings1, pdImplicitArgUsed=implicitArgUsed, pdNextUniqLetId=letid }
unionUsed lhs rhs =
   let (newOnes,oldOnes) = splitAt (length rhs - length lhs) rhs
   in newOnes ++ (zipWith (\a b->b { argsImpls=zipWith unionUsed1 (argsImpls a) (argsImpls b) } ) lhs oldOnes)
   where unionUsed1 lhs rhs =
            rhs { implUsed = if implUsed rhs==UsedArg then implUsed rhs else implUsed lhs }
data ArgAttemptResult =
   Success
      -> ParseState ArgAttemptResult
tryArg (Cond desc c) _ prevTypes nibs memoArgs = do
   let (impl,nextState) = head memoArgs
   state <- get
   code <- gets pdCode
   if c (MatchTestData (prevTypes++[implType impl]) nibs state)
   then do
      putAddRep nextState
      return $ Success (tail memoArgs) [impl]
   else return $ FailTypeMismatch $ "arg " ++ show (length prevTypes+1) ++ " is " ++ show (implType impl)
tryArg (ParseArg _ parser) _ _ _ _ = do
   (t,code) <- parser
   return $ Success (error"todo memo args") [noArgsUsed { implType=t, implCode=hsAtom code }]
tryArg Auto _ _ _ memoArgs = do
   matched <- match tildaOp
   return $ if matched then Success (tail memoArgs) [] else FailTypeMismatch "arg isn't ~"
tryArg (FakeAuto _) _ _ _ memoArgs = do
   return $ Success memoArgs []
tryArg (BinCode b) _ _ _ memoArgs = do
   matched <- match ([b], [])
   stateThunk <- toThunk
   code <- gets pdCode
   let regenedArgs = head $ exprsByOffset stateThunk
   return $ if matched then Success (if isBinary code then regenedArgs else memoArgs) [] else FailTypeMismatch "arg bincode mismatch"
tryArg NotEOF _ _ _ memoArgs = do
   code <- gets pdCode
   implicitArgUsed <- gets pdImplicitArgUsed
   return $ if not (isBinary code) && empty code && implicitArgUsed
      then FailTypeMismatch "EOF cannot be here since this op needs part of its binary representation after the arg (try not using implicit args)"
      else Success memoArgs []
tryArg EOF _ _ _ memoArgs = do
   code <- gets pdCode
   return $ if empty code
      then Success memoArgs []
      else FailTypeMismatch "not EOF"
tryArg (LitCode l) _ _ _ memoArgs = do
   matched <- match ([], [[l]])
   stateThunk <- toThunk
   code <- gets pdCode
   let regenedArgs = head $ exprsByOffset stateThunk
   return $ if matched then Success (if not $ isBinary code then regenedArgs else memoArgs) [] else FailTypeMismatch "arg litcode mismatch"
tryArg AnyS from prevTs _ _ =
   tryArg (Fn ReqDontCare UnusedArg (const (1,[]))) from prevTs undefined undefined
tryArg (Fn reqArgUse argUsedness f) from prevTs _ _ = do
   let (nRets, argT) = f prevTs
   (impl,used) <- getLambdaValue nRets argT argUsedness from (reqArgUse /= ReqDontCare)
   let success = Success (error"memoized args cannot be used after fn")
   return $ case reqArgUse of
      ReqArg | not (or used) && not ?isSimple -> FailConstFn impl
      ReqConst -> if (or used) && not ?isSimple
         then FailTypeMismatch $ "arg " ++ show (length prevTs+1) ++ " is not a const"
         else success [app1Hs (fillAccums (length argT) (length argT)) impl]
      otherwise -> success [impl]
tryArg (OptionalFn f) from prevTs _ memoArgs = do
   let (nRets, argT) = f prevTs
   (impl,used) <- getLambdaValue nRets argT UnusedArg from False
   code <- gets pdCode
   context <- gets pdContext
   nextUniqLetId <- gets pdNextUniqLetId
   if not (or used) then
      return $ Success ((impl, blankRep code context nextUniqLetId):error "cannot used memo args after the one after optional fn") [noArgsUsed { implType=ItWasAConstant}]
   else
      return $ Success (head $ exprsByOffset $ Thunk code context nextUniqLetId) [impl]
tryArg (OrAuto _ nonAutoSpec) from prevTs nibs memoArgs = do
   matched <- match tildaOp
   if matched
   then return $ Success (tail memoArgs) [noArgsUsed { implType=OptionYes }]
   else tryArg nonAutoSpec from prevTs nibs memoArgs
tryArg (AutoNot fn) from prevTs _ _ = do
   matched <- match tildaOp
   afterArg <- tryArg fn from prevTs (error"impossible 43") (error"impossible 44")
   return $ case afterArg of
      Success memo [impl] ->
         let
            truthyImpl = app1Hs ((truthy $ ret $ implType impl)++".") impl
            modifiedImpl = if matched then app1Hs "not." truthyImpl else truthyImpl
         in Success memo [modifiedImpl]
tryArg (AutoOption desc) _ prevTs nibs memoArgs = do
   matched <- match tildaOp
   return $ if matched
      then Success (tail memoArgs) [noArgsUsed { implType=OptionYes }]
      else Success       memoArgs  [noArgsUsed { implType=OptionNo }]
tryArg (AutoDefault tspec v) from prevTypes nibs memoArgs = do
   matched <- match tildaOp
   if matched
   then return $ Success (tail memoArgs) [noArgsUsed { implType=VInt, implCode=i v }]
   else tryArg tspec from prevTypes nibs memoArgs
tryArg (AutoData tspec) from prevTypes nibs memoArgs = do
   matched <- match tildaOp
   if matched
   then do
      modify $ \s -> s { pdDataUsed = True }
      return $ Success (tail memoArgs) [noArgsUsed { implType=VInt, implCode=hsAtom"dat" }]
   else tryArg tspec from prevTypes nibs memoArgs
tryArg CharClassMode _ _ _ memoArgs = do
   impl <- parse1Nibble "char class mode" $ zip [0..] charClasses
   return $ Success (error"memoized args cannot be used after char class mode (but could be)") [impl]
tryArg ZipMode _ [t1,VFn _ [t2]] _ memoArgs = do
   impl <- parse1Nibble "zip mode" $ zip [0..] (specialZips [t1,t2])
   return $ Success (error"memoized args cannot be used after zip mode (but could be)") [impl]
tryArg FoldMode _ prevTypes _ memoArgs = do
   impl <- parse1Nibble "fold mode" $ zip [0..] (specialFolds prevTypes)
   return $ Success (error"memoized args cannot be used after fold mode (but could be)") [impl]
createImplMonad t hs = return $ noArgsUsed { implType=t, implCode=hs }
createSpecialFn (ts,hs) = createImplMonad (VFn undefined ts) (hsParen $ hsAtom hs)
createZipFromBinOp c a@[a1,a2] lhsNeed rhsNeed =
   let lhsDim = cidim [a1]
       rhsDim = cidim [a2]
       (t,hs)=binOp c [head $ iterate ((:[]).VList) [baseElem a1] !! lhsNeed,baseElem a2]
       (vec,extra) = fullVectorize (lhsDim-lhsNeed) (rhsDim-rhsNeed)
       vt = iterate ((:[]).VList) t !! extra
   in createImplMonad (VFn undefined vt) (hsParen $ hsAtom $ vec ++ "(" ++ hs ++ ")")
specialZips a@[a1,a2] = let
      (a2p,ap2Fn) = promoteListRepeat [a2]
      [l1,l2] = map (length.elemT) [a1,a2p]
      flatten = flattenTuples l1 l2
      flatT = concatMap elemT [a1,a2p]
      (a1p,ap1Fn) = promoteList (elemT a1)
      (coercedType, coerceFnA, coerceFnB) = coerce [a1p] (elemT a2p)
   in
      [('~',getLambdaValue 1 flatT UnusedArg "zip with by" False >>= \(impl,_) -> return $ app1Hs  ("(zipWith . \\f a b->f $ "++flatten++"(a,b))") impl { implType=VFn undefined [VList $ ret $ implType impl] })
      ,(':', createSpecialFn ([VList coercedType], "\\a1 b1->zipWith (\\a b->("++coerceFnA++"$ "++ap1Fn++"a)++"++coerceFnB++"b) a1 ("++ap2Fn++"b1)"))
      ,(',',createSpecialFn ([VList flatT], "\\a1 b1->zipWith (\\a b->"++flatten++" $ (a,b)) a1 ("++ap2Fn++"b1)"))
      map (\c->(c,createZipFromBinOp c a 0 0)) "+*-/%^][!" ++
      map (\c->(c,createZipFromBinOp c a 1 0)) "=?"
createFoldFromBinOp (c,initValue) (VList [t]) =
   let (vt,hs)=binOp c [t,t]
   in createImplMonad (VFn undefined vt) (hsParen $ hsAtom $ "\\(foldType,initFn) a->if null a then initFn $ "++initValue++" else foldType (" ++ hs ++ ") a")
specialFolds [a1] =
   (map (\deets->(fst deets,createFoldFromBinOp deets a1))
      [(']',minForType $ elemT a1)
      ,('[',maxForType $ elemT a1)
      ,('|',defaultValue $ elemT a1)
      ,('&',maxForType $ elemT a1)
      ,('+',"0")
      ,('*',"1")
      ,('-',"0")
      ,('/',"1")
      ,('%',"1")
      ,('^',"1")])++
      [('>', takeAnotherOrderBy ">")
      ,('<', takeAnotherOrderBy "<")
      ,(':', createImplMonad (VFn undefined [a1]) (hsAtom $ "\\foldType a -> tail $ inits a"))
      ,(';', createImplMonad (VFn undefined [a1]) (hsAtom $ "\\foldType a -> init $ tails a"))
      ] where
   takeAnotherOrderBy fName = do
      getLambdaValue 1 (elemT a1) UnusedArg ("special fold "++fName) False >>= (\(impl,_) -> return $ app1Hs ("(\\f (foldType,initFn) a->if null a then initFn $ "++defaultValue (elemT a1)++" else foldType (onToSelectBy ("++fName++") f) a)") impl { implType=VFn undefined (elemT a1) } )
charClasses = map (\(c,hs)->(c,createImplMonad (VFn undefined [VInt]) (hsParen $ hsAtom hs))) charClassesDefs
charClassesDefs =
   ,('A',"not.isAlpha")
   ,('n',"isAlphaNum")
   ,('N',"not.isAlphaNum")
   ,('s',"isSpace")
   ,('S',"not.isSpace")
   ,('l',"isLower")
   ,('L',"not.isLower")
   ,('u',"isUpper")
   ,('U',"not.isUpper")
   ,('d',"isDigit")
   ,('D',"not.isDigit")
   ,('$',"isSym")
   ,('!',"not.isSym")]
addEofChecks (a:BinCode c:as) = a:NotEOF:BinCode c:addEofChecks as
addEofChecks (a:as) = a:addEofChecks as
addEofChecks [] = []
getArgs argSpecs = getArgsH [] [] (addEofChecks argSpecs)
getArgsH prevArgs _ [] from _ = return $ Left prevArgs
getArgsH prevArgs prevNibs (spec:s) from memoArgs = do
   arg <- tryArg spec from prevArgTypes nibs memoArgs
   case arg of
      FailConstFn _ -> return $ Right "lambda fn does not use its argument but is required to"
      FailTypeMismatch reason -> return $ Right reason
      Success nextMemoizedArgs impl -> getArgsH (prevArgs++impl) nibs s from nextMemoizedArgs
   where
      afterState = snd $ head memoArgs
      nibs = (prevNibs++if pdImplicitArgUsed afterState then [newSmartList []] else [theseNibs])
      prevArgTypes = map implType prevArgs
getLambdaValue numRets argType argUsedness from reqUsed = do
   (newArg,body) <- pushLambdaArg argType argUsedness from $ \newArg checkIfArgsUsedFn -> do
      if numRets == 0 then do
         origDepth <- gets pdContext >>= return.length
         let nonRecImpls = init $ argsImpls newArg
         modify $ \s -> s { pdContext=Args nonRecImpls LambdaArg "recursive fn" : tail (pdContext s) }
         match <- match tildaOp
         aa <- get1Value
         let truthyF = truthy [implType aa]
         let a = app1Hs ((if match then "not$" else "")++truthyF) aa { implType = InvalidType }
         bonus <- parseCountTuple
         b <- getValuesMemo (1+bonus)
         let from = map implType nonRecImpls
         let toType = map implType b
         let recType = VFn from toType
         let recImpl = (last $ argsImpls newArg) { implType=recType }
         let recArg = Args (nonRecImpls ++ [recImpl]) LambdaArg "recursive fn"
         modify $ \s -> s { pdContext = replaceTailOfIth (newDepth-origDepth) (pdContext s) recImpl }
         c <- getNArgs toType
         return $ [a]++[makePairs argType b]++[makePairs argType c]
      else do
         bonus <- parseCountTuple
         if reqUsed && bonus == 0 then
            getValuesMemo2 numRets $ \context -> checkIfArgsUsedFn context
         else
            getValuesMemo (bonus + numRets)
   return $ (addLambda newArg body, map (\impl->UsedArg==implUsed impl) $ argsImpls newArg)
replaceTailOfIth n context new = zipWith (\i arg->
      if i==n then arg { argsImpls = (argsImpls arg) ++ [new] }
      else arg
   ) [0..] context
pushLambdaArg argType argUsedness from f = do
   code <- gets pdCode
   namedArgs <- if not (null argType) then case onlyCheckMatch code ([-1],["\\"]) of
      Just nextCode -> do
         maybeName <- getNewIdentifier nextCode
         case maybeName of
            Left (name,nextCode2) -> do
               modify $ \s -> s { pdCode = nextCode2 }
               rest <- consumeNIdentifiers (length argType-1) $ "lambda (expecting "++(show$length argType)++" args)"
               return $ Just $ name : rest
            Right reason | reason == inUseAlreadyMsg -> do
               parseLitWarning $ "You may have meant to use a lambda failed because " ++ reason
               return Nothing
   newArg <- newLambdaArg argType namedArgs argUsedness from
   depth <- gets pdContext >>= return.length
   rets <- f newArg (\context -> (any ((==UsedArg).implUsed) $ argsImpls $ getNewArg depth context))
   let body = makePairs argType rets
   finalContext <- gets pdContext
   let newArgFinal = getNewArg depth finalContext
   bodyWithLets <- popArg depth body
   return (newArgFinal, bodyWithLets)
consumeNIdentifiers 0 _ = return []
consumeNIdentifiers n exprType = do
   code <- gets pdCode
   maybeName <- getNewIdentifier code
   case maybeName of
      Left (name, nextCode) -> do
         modify $ \s -> s { pdCode = nextCode }
         rest <- consumeNIdentifiers (n-1) exprType
         return $ name:rest
      Right reason -> parseError $ "invalid " ++ exprType ++ ": " ++ reason
inUseAlreadyMsg = "identifier is already in use"
getNewIdentifier code = do
   allArgs <- getAllArgs
   let maybeName = onlyCheckMatchIdentifier code
   return $ case maybeName of
      Just (name, nextCode) ->
         if not (any (==Just name) $ map implName allArgs)
           && notElem name opsWithIdentiferNames then
            Left (name, nextCode)
         else Right inUseAlreadyMsg
      otherwise -> Right "not an identifier"
   where opsWithIdentiferNames = nub $ filter (all isIdentifierChar) $ flip map allOps (\(_,lit,_,_)->concat lit)
get1Value = do
   v <- getValuesMemo 1
   return $ head v
toThunk = do
   state <- get
   return $ Thunk (pdCode state) (pdContext state) (pdNextUniqLetId state)
getValuesMemo n = do
   stateThunk <- toThunk
   let exprs = take n $ head $ exprsByOffset stateThunk
   mapM (putAddRep.snd) exprs
   return $ map fst exprs
getValuesMemo2 n moreIfFn = do
   stateThunk <- toThunk
   let exprs = take n $ head $ exprsByOffset stateThunk
   let more = moreIfFn $ pdContext $ snd $ head exprs
   let finalExprs = if more then exprs else [head exprs]
   mapM (putAddRep.snd) finalExprs
   return $ map fst finalExprs
data Thunk = Thunk Code [Args] Int
getValues (Thunk code context nextUniqLetId) offsetExprs = (impl, after) : getValues afterThunk offsetAfterExprs where
   (impl, after) = runState (do
      didThem <- doLitLets
      case didThem of
         Just nextGetValue -> nextGetValue
         Nothing -> getValue offsetExprs) $ blankRep code context nextUniqLetId
   afterThunk = Thunk (pdCode after) (pdContext after) (pdNextUniqLetId after)
   offsetAfterExprs =
      if pdNextUniqLetId after == nextUniqLetId
         && equivalentContexts (pdContext after) context
      then drop (cp (pdCode after) - cp code) offsetExprs
      else drop 1 $ exprsByOffset afterThunk
equivalentContexts a1s a2s =
   length a1s == length a2s
   && and (zipWith equallyUsed (concatMap argsImpls a1s) (concatMap argsImpls a2s))
   where
      equallyUsed a1 a2 = implUsed a1 == implUsed a2
exprsByOffset thunk@(Thunk code context nextUniqLetId) =
   getValues thunk rest : rest where
      rest = exprsByOffset (Thunk (nextOffset code) context nextUniqLetId)
data FailedMatch = FailedMatch {
}
getValue memoArgOffsets = do
   let ops = if ?isSimple then simpleOps else allOps
   code <- gets pdCode
   identifier <- getArgByName
   if empty code
   then argImplicit
   else if isJust $ identifier then applyFirstClassFn (fromMaybe undefined identifier)
   else getValueH ops memoArgOffsets []
getValueH [] _ failedMatches = parseError $ "Parse Error: no matching op" ++ if null attempts then " (no matching identifier)" else ", tried: " ++ attempts
   where attempts = flip concatMap failedMatches $
            \(FailedMatch opName matchFailureReason expectedTypes) ->
               "\n"++opName++" "++unwords expectedTypes++", "++matchFailureReason
getValueH ((isPriority,lit,nib,op):otherOps) memoArgOffsets failedMatches = do
   code <- gets pdCode
   let tryRest newFailedMatches = getValueH otherOps memoArgOffsets newFailedMatches
   let reconstructedLit = dToList $ pdLit $ snd $ head $ head memoArgOffsets
   origState <- get
   match <- match (nib,lit)
   if not match then tryRest failedMatches
   else if not isPriority && isBinary code && concat lit !! 1 /= head reconstructedLit
   then do
      put origState
      tryRest failedMatches
   else do
      afterOpCode <- gets pdCode
      let valList = head (drop (cp afterOpCode - cp code - 1) memoArgOffsets)
      attempt <- convertOp valList (concat lit) op code
      case attempt of
         OpMatch impl -> return impl
         OpLitWarn msg -> do
            put origState
            parseLitWarning msg
            tryRest failedMatches
         OpTypeMismatch reason -> do
            put origState
            let opName = concat lit
                expectedTypes = catMaybes $ map typeToStr $ fst op
            tryRest $ (FailedMatch opName reason expectedTypes) : failedMatches
data OpMatchResult = OpMatch Impl | OpTypeMismatch String | OpLitWarn String
convertOp memoizedArgList opName (ats,behavior) preOpCode = do
   maybeArgs <- getArgs ats opName memoizedArgList
   case (maybeArgs, behavior) of
      (Right reason,_) -> return $ OpTypeMismatch reason
      (Left args, LitWarn msg) -> do
         return $ OpLitWarn msg
      (Left args, CodeGen impl) -> do
         afterArgsCode <- gets pdCode
         modify $ \s -> s { pdCode=preOpCode }
         parseCoordinates <- getParseCoordinates
         (rt, initImpl) <- impl $ map implType args
         afterImplCode <- gets pdCode
         if (cp preOpCode) /= (cp afterImplCode)
         then error "modifying code in op impl is not supported (but it could be)"
         else modify $ \s -> s { pdCode=afterArgsCode }
         let fullImpl = foldl applyImpl initImpl args
         unpairedFirst <- convertPairToLet (implUsed initImpl) fullImpl rt $ opName ++ " " ++ parseCoordinates
         result <- applyFirstClassFn unpairedFirst
         return $ OpMatch result
applyFirstClassFn (Impl (VFn from to) hs dep _ _) = getNArgs from >>= \impls -> do
   let initImpl = Impl undefined hs dep undefined undefined
   convertPairToLet UnusedArg (foldl applyImpl (app1Hs (curryN (length impls)) initImpl) impls) to "first class fn application"
applyFirstClassFn x = return x
getNArgs argTypes = do
   args <- getValuesMemo (length argTypes)
   return $ zipWith coerceImpl args argTypes
coerceImpl (Impl et hs dep _ _) t = Impl t (hsApp (hsAtom$coerceTo [t] [et]) hs) dep undefined undefined
convertPairToLet _ (Impl _ hs dep _ _) [t] _ = return $ Impl t hs dep undefined undefined
convertPairToLet argUsedness impl implTypes from = do
   setVars <- getSetVars (length implTypes-1)
   letArg <- newLetArg setVars argUsedness impl implTypes from
   return $ head $ argsImpls letArg
getSetVars n = do
   code <- gets pdCode
   case onlyCheckMatch code ([-1], ["sets "]) of
      Just nextCode -> do
         modify $ \s -> s { pdCode=nextCode }
         consumeNIdentifiers n "sets expression" >>= (\names-> return $ Just $ "first is unused" : names)
      otherwise -> return Nothing
getParseCoordinates = do
   c <- gets pdCode
   return $ case c of
      (Nib _ _) -> ""
      (Lit f _ cp) ->
         let (lineno,charno) = getLitParseCoordinates f cp in
            "(line:"++show lineno++",char:"++show charno++")"
doLitLets = do
   code <- gets pdCode
   if matched then do
      name <- consumeNIdentifiers 1 "let expression"
      impl <- get1Value
      let doubledImpl = app1Hs "(\\x->((),x))" impl
      newLetArg (Just $ "first is unused" : name) OptionalArg doubledImpl [implType impl,implType impl] "let"
      return (Just get1Value)
version = "nibbles 0.25+ (unstable)"
webpage = "http://golfscript.com/nibbles"
data ArgUsedness = UnusedArg | OptionalArg | UsedArg | UsednessDoesntMatter deriving (Show, Eq)
                 } deriving (Eq, Show)
noArgsUsed = Impl (error "undefined impl type") (hsAtom "(error \"undefined impl code\")") (Set.singleton 0) Nothing UnusedArg
isBinary (Nib _ _) = True
isBinary _ = False
app1Hs s impl = impl { implCode=hsApp (hsAtom s) (implCode impl) }
type ParseState = State ParseData
data ReqArg = ReqDontCare | ReqArg | ReqConst deriving Eq
data ArgSpec
   = Cond String {-desc-} (MatchTestData -> Bool)
   | ParseArg String {-desc-} (ParseState (VT, String))
   | Auto
   | AutoData ArgSpec
   | Fn ReqArg ArgUsedness ([VT] -> (Int, [VT]))
   | OrAuto String {- desc -} ArgSpec
   | AutoNot ArgSpec {- only Fn -}
   | AutoOption String {- desc -}
   | ZipMode
   | FoldMode
   | CharClassMode
   | BinCode Int
   | LitCode Char
   | NotEOF
   | EOF
data OpBehavior = LitWarn String | CodeGen ([VT]->ParseState ([VT], Impl))
type Operation = ([ArgSpec], OpBehavior)
dToList = DList.toList
addLitWarning msg = do
   modify $ \s -> s { pdLitWarnings = msg : pdLitWarnings s }
   return ()
appendRepH (nib2,lit2) = do
   nib <- gets pdNib
   lit <- gets pdLit
   modify $ \s -> s { pdNib = smartAppend nib nib2
                    , pdLit = DList.append lit lit2 }
appendRep (nib2,lit2) = appendRepH (newSmartList nib2, DList.fromList lit2)
appendRepA (nib,lit) = appendRep (nib,concat lit)
blankRep code context nextUniqLetId =
   ParseData code context (newSmartList []) (DList.fromList "") False False [] nextUniqLetId
getNib = smartToList . pdNib
getLit = DList.toList . pdLit
litFile = quoteFile $ QuasiQuoter {quoteExp = return . LitE . StringL,
                                   quotePat = undefined,
                                   quoteType = undefined,
                                   quoteDec = undefined}
newli = myOrd '\n'
space = myOrd ' '
printables = sToA $ ' ':['a'..'z']++".,!?_\n"++['A'..'Z']++['0'..'9']
   ++"-+:;\"'~`@#$%^&*()[]{}<>\\/=|"
unprintables = filter (\c->not$elem c printables) $ sToA ['\0'..'\127']
myChr i = chr $ fromIntegral i
myOrd c = fromIntegral $ ord c
sToA = map (fromIntegral.myOrd)
aToS = map$myChr.fromIntegral
bToI b = if b then 1 else 0
onToBy f x y = f x == f y
onToSelectBy op f x y = if op (f x) (f y) then x else y
safeDiv a 0 = signum a*(2^128)
safeDiv a b = a `div` b
safeMod a 0 = 0
safeMod a b = a `mod` b
safeDivMod a b = (safeDiv a b, safeMod a b)
confirmInt = id
confirmList = id
at [] _ = Nothing
at (x:xs) n
 | n < 0     = Nothing
 | n == 0    = Just x
 | otherwise = at xs (n-1)
step n a = map fst $ filter ((==0).(`mod`n).snd) (zip (if n<0 then reverse a else a) [0..])
iff c b1 b2 = if c then Left b1 else Right b2
finishLn = finishLnHelper True
finishLnHelper _ (x:xs) = x:finishLnHelper (x==newli) xs
finishLnHelper True [] = []
finishLnHelper False [] = [newli]
lazyAtMod a i = fromMaybe
 (a!!fromIntegral(i`mod`fromIntegral (length a)))
 (at a i)
asIntsH "" = []
asIntsH "-" = []
asIntsH ('-':c:rest)
   | isDigit c = ('-':c:num) : asIntsH after
   | otherwise = asIntsH rest
   where
      (num,after) = span isDigit rest
asIntsH (c:rest)
   | isDigit c = (c:num) : asIntsH after
   | otherwise = asIntsH rest
   where
      (num,after) = span isDigit rest
toBase b n = reverse $ map (fromIntegral.flip mod b) $ takeWhile (>0) $ iterate (flip div b) (fromIntegral n)
hlist a = fromIntegral $ Murmur.asWord64 $ Murmur.hash64 $ B8.pack $ map fromIntegral $ concatMap (toBase 256)
    (map fromIntegral a)
listOr defaultResult [] = defaultResult
listOr _ nonEmpty = nonEmpty
parseNum base strI = if base > 36 then error "parseNum base must be <= 36" else
   case findIndex isJust digitIndices of
      Just i -> let value = fromBase base $ catMaybes $ takeWhile isJust $ drop i digitIndices
                    sign = if isSuffixOf "-" (take i str) then -1 else 1 in
                    sign * value
      otherwise -> 0
   where
      str = map toLower $ aToS strI
      digits = genericTake base $ ['0'..'9']++['a'..'z']
      digitIndices = map (flip elemIndex digits) str
appendUntilNull a f = a ++ appendUntilNullH (reverse a) f
appendUntilNullH a f = case f a of
   [] -> []
   r -> r ++ appendUntilNullH (reverse r++a) f
iterateWhileUniq f i =
   let uniqPart = iterateWhileUniqH Set.empty f i
       Just repeatedIndex = elemIndex (f (last uniqPart)) uniqPart
iterateWhileUniqH been f i =
   if Set.member i been then []
   else i : iterateWhileUniqH (Set.insert i been) f (f i)
subsequencesN _ []     = [[]]
subsequencesN 0 _      = [[]]
subsequencesN n xs | n >= genericLength xs = [xs]
subsequencesN n (x:xs) = (map (x:) $ subsequencesN (n-1) xs) ++ subsequencesN n xs
repeatedSubsequencesN _ []     = []
repeatedSubsequencesN 1 xs      = map (:[]) xs
repeatedSubsequencesN n (x:xs) = (map (x:) $ repeatedSubsequencesN (n-1) (x:xs)) ++ repeatedSubsequencesN n xs
nthRoot n 0 = 0
nthRoot n x = nthRootGuess n x x
nthRootGuess n guess x = case ((x `div` guess^(n-1)) + guess*(n-1)) `div` n of
   r | r>=guess -> guess
     | otherwise -> nthRootGuess n r x
permutationsSaneOrder [] = [[]]
permutationsSaneOrder as = do
   a <- as
   let l = delete a as
   ls <- permutationsSaneOrder l
   return $ a : ls
isSym c = isPunctuation c || isSymbol c
toHex = sToA . flip showHex ""
single = singleton
flist = fromList
a +++ b = append a b
newtype HsCode = HsCode (DList Char) deriving (Show, Eq)
getHsCode (HsCode hs) = hs
hsAtom = HsCode . fromList
hsApp (HsCode a) (HsCode b) = hsParen $ HsCode $ a +++ single ' ' +++ b
hsFn args (HsCode body) = HsCode $ flist "(\\" +++ argsLhs args +++ flist"->" +++ body +++ single ')' where
   argsLhs hss = getHsCode $ hsParen $ HsCode $ intercalate (single ',') $ Prelude.map getHsCode hss
hsLet vars (HsCode def) (HsCode body) =
   HsCode $ flist"(let (" +++ lhs +++ flist")="
      +++ def +++ flist" in "
      +++ body +++ single ')'
         where lhs = intercalate (single ',') $ Prelude.map getHsCode vars
hsParen (HsCode hs) = HsCode $ single '(' +++ hs +++ single ')'
flatHs (HsCode hs) = toList hs
tildaRep = (head $ head $ snd tildaOp, head $ fst tildaOp)
intRep = ('\0', 1)
strRep = ('"',2)
setRep = (';',6)
rawOps = [
   op(tildaRep, [], "error\"undefined auto impl\""~>AutoType),
   op(intRep, [ParseArg "int" intParser], ()),
   op(strRep, [ParseArg "str" strParser], ()),
   extendOp "'" [lengthRep, strRep] (shorterReason"just hardcode the length of the string") ([ParseArg "chr" chrParser], ()),
   op(('$', 3), [], argn 1),
   op(('@', 4), [], argn 2),
   op(('_', 5), [], argn 3),
   op(setRep, [any1], "\\x->(x,x)" ~> dup.a1),
   op(consRep, [AnyS, auto], "\\v->v():[]" ~> VList .ret.a1),
   op(consRep, [AnyS, FakeAuto "[]", AnyS], \[a,b]->
      let
         (ap,apFn) = promoteList (ret a)
         (bp,bpFn) = promoteList (ret b)
         (coercedType, coerceFnA, coerceFnB) = coerce [ap] [bp]
      in
         "\\a b->("++coerceFnA++"$"++apFn++"(a()))++("++coerceFnB++"$"++bpFn++"(b()))"~>coercedType
      ),
   extendOp "`/" [subscriptRep] (shorterReason"this could be accomplished with mod after length") ([num, BinCodeRep rangeRep,AutoDefault num 2], "safeDivMod" ~> [VInt, VInt]),
   extendOp "`%" [takeRep] (shorterReason"this could be done by doing a range on the min of the two num values") ([num,BinCodeRep rangeRep,AutoDefault num 2], "(swap.).safeDivMod" ~> [VInt,VInt]),
   commutativeExtension [8]
      ("]", [AutoDefault num 0, num], "max"~>orChr)
      ("+", [AutoDefault num 1, AutoDefault vec 1], vectorize "+" xorChr),
   commutativeExtension [10]
      ("[", [int, num], "min"~>orChr)
      ("*", [AutoDefault int (-1), AutoDefault vec 2], vectorize "*" (const VInt)),
   op(subtractRep, [AutoDefault num 1, AutoDefault num 1], "-" ~> xorChr),
   op(('/',11), [num, AutoDefault num 2], "safeDiv" ~> VInt),
   op(('%',12), [num, AutoDefault num 2], "safeMod" ~> VInt),
   op(('^',14), [AutoDefault int 10, AutoDefault int 2], "\\a b->if b<0 then nthRoot (-b) a else a^b" ~> a1),
   op(sumRep, [listOf int], \[a1]->
      let (uzT,uzF)=unzipTuple a1 in
         appFst uzT "sum" ++ "." ++ uzF ~> VInt : tail uzT
      ),
   op(sumRep, [listOf list], \[a1]->let (uzT,uzF)=unzipTuple a1 in
         appFst uzT "concat" ++ "." ++ uzF ~> head (elemT (head uzT)) : tail uzT
      ),
   extendOpM ["=","~"] [lengthRep,addRep] (shorterReason"length of vec+ is just length of") ([num, list], \[a1,a2]->"\\i a->fromMaybe "++defaultValue (elemT a2)++" $ at a (fromIntegral i - 1)" ~> elemT a2),
   op(mapRep, [list, fn (elemT.a1)], mapFn ~> VList .ret.a2),
   op(filterRep, [list, auto, AutoNot $ fn (elemT.a1)], "\\a f->filter (not.f) a" ~> a1),
   op(filterRep, [list, AutoNot $ Fn ReqArg UnusedArg $ \[a1]->(1,elemT a1)], "\\a f->filter f a" ~> a1),
   op(('!',snd filterRep), [list, Fn ReqConst UnusedArg $ \[a1]->(1,elemT a1), ZipMode], \[a1,a2,rt]->"\\a b f->f a b" ~> ret rt),
   op(('/',10), [list, auto, fn2 (const []), fnx (\[a1,a2]->(length $ ret a2, elemT a1 ++ ret a2))], foldrFn "foldr" ~> ret.a2),
   op(('/',10), [nonTupleList, BinCode 3, LitCode '$'], \[a1] -> "\\a -> if null a then "++defaultValue (elemT a1)++" else head a" ~> head (elemT a1)),
   op(('/',10), [list, EOF], \[a1] -> do
      modify $ \s -> s { pdImplicitArgUsed = True }
      return $ "\\a -> if null a then "++defaultValue (elemT a1)++" else head a" ~> elemT a1),
   op(('/',10), [list, Fn ReqArg UnusedArg $ \[a1]->(length $ elemT a1,concat $ replicate 2 $ elemT a1)], foldr1Fn "foldr1" "id" ~> elemT.a1),
   op(('/',10), [list, FakeAuto $ italic "tuple", Fn ReqConst UnusedArg $ \[a1]->(length $ elemT a1,concat $ replicate 2 $ elemT a1), fn (\[a1,a2]->elemT a1 ++ ret a2)], foldrFn2 "foldr" ~> ret.a2),
   op(reverseRep, [list], "reverse" ~> a1),
   op(lengthRep, [list], "genericLength" ~> VInt),
   op(takeRep, [FakeAuto "while", num, list], "\\n a->genericTake (if n<0 then genericLength a+n else n) a" ~> a2),
   op(dropRep, [FakeAuto "while", int, list], "\\n a->genericDrop (if n<0 then genericLength a+n else n) a" ~> a2),
   extendOp "`," [reverseRep, consRep] equivalentOrderReason ([AutoDefault num (2^128)], "\\x->[0..x-1]" ~> vList1 . a1),
   op(rangeRep, [AutoDefault num (2^128)], "\\x->[1..x]" ~> vList1 .a1),
   op(repRep, [AutoDefault int (2^128), orC list char], \[_,a2] ->
      let (ap2, apf) = promoteList [a2] in
      "\\n a->concat$genericReplicate n$"++apf++" a" ~> ap2),
   op(subscriptRep, [FakeAuto $ italic "nowrap", num, list], \[a1,a2]->"\\i a->if null a then "++defaultValue (elemT a2)++" else lazyAtMod a (fromIntegral i - 1)" ~> elemT a2),
   op(('?',15), [list, auto, AutoNot $ fn (elemT.a1)], "\\l f->fromIntegral$1+(fromMaybe (-1) $ findIndex f l)" ~> VInt),
   op(('?',15), [listToBeReferenced, FakeAuto "by", elemOfA1], \[a1,a2]->"\\a e->fromIntegral$1+(fromMaybe (-1) $ elemIndex e (map "++firstOf (elemT a1)++" a))" ~> VInt),
   op(diffRep, [listToBeReferenced, sameAsA1], "\\\\" ~> a1),
   op(('%',8), [str, OrAuto "words" $ orC str char], \[a1,a2]->
      (if a2==OptionYes then "\\a->map sToA $ words (aToS a)" else
      let (ap2, apf) = promoteList [a2]
      in "(\\a b->filter (/=[])$splitOn ("++apf++"b) a)") ~> vList1 a1),
   op(('*',8), [str, list], Polylib.join.a2),
   op(('&',8), [str, int, AutoOption "center", vec], \[a1,a2,o1,a3] ->
      let justify t =
           "\\s n o->let oc = "++coerceTo [vstr] [t]++"o ;\
                      \pad = genericTake (abs n-genericLength oc) (if null s then s else cycle s) in " ++
           (if o1 == OptionYes then
               "let (smallPad,bigPad)=splitAt (length pad `div` 2) pad in \
               \if n >= 0 then smallPad ++ oc ++ bigPad else bigPad ++ oc ++ smallPad"
            else
               "if n >= 0 then pad ++ oc else oc ++ pad") in
      if isList a3 && a3 /= vstr then
         "\\s n o->let ocs = map "++coerceTo [vstr] (elemT a3)++" o;\
                     \ nmax = (if n<0 then -1 else 1) * (maximum $ abs n : map genericLength ocs) in \
                     \ map ((" ++ justify vstr ++ ")s nmax) ocs"  ~> vList1 vstr
      else justify a3 ~> vstr),
   op(('\\',10), [char, CharClassMode], "\\a f->if f $ myChr a then [a] else []" ~> vstr),
   op(('o',14), [char], "id"~>VInt),
   opM(["`r"], [15], [str, BinCode 1], "parseNum 10" ~> VInt),
   lowPriorityOp([fst ifRep,fst lengthRep], [snd ifRep,snd lengthRep], list:ifBodyArgs, ifImpl),
   op(ifRep, num:ifBodyArgs, ifImpl),
   opM([">",">"], [12,0] {- if change, then change hidden warning for map on tail -}, [list], "tail" ~> a1),
   opM(["<","<"], [11,0 {- if change then change hex -}], [list], "init" ~> a1),
   opM(["`","("], [14], [list, BinCode 1], \[a1]->flattenTuples (length$elemT a1) 1++".fromMaybe("++defaultValue(elemT a1)++",[]).uncons" ~> elemT a1++[a1]),
   opM(["`",")"], [14], [list, BinCode 2], \[a1]->flattenTuples 1 (length$elemT a1)++".swap.fromMaybe("++defaultValue(elemT a1)++",[]).uncons" ~> a1:elemT a1),
   extendOp "`/" [reverseRep,repRep] equivalentOrderReason ([AutoDefault num 2, list], "\\n a->if n<0 \
   \then reverse $ map reverse $ chunksOf (fromIntegral (-n)) (reverse a)\
   \else chunksOf (fromIntegral n) a" ~> vList1 .a2),
   extendOp "`\\" [lengthRep, repRep] "Multiply length by fst arg instead of replicating list." ([AutoDefault int 2, list], "\\a b->map (map fst) $ groupBy (onToBy $ \\e->a*(snd e + if a<0 then 1 else 0)`div`genericLength b) $ zip b [0..]" ~> vList1 . a2),
   extendOp "`%" [takeRep, mapRep] equivalentOrderReason ([AutoDefault num 2, list], "step" ~> a2),
   opM("`?", [14], [list, BinCode 10, AutoNot $ Fn ReqArg UnusedArg $ \[a1]->(1,elemT a1)], "\\l f->map ((+1).fromIntegral) $ findIndices f l" ~> vList1 VInt),
   opM("`?", [14], [list, BinCode 10, AutoOption "not", Fn ReqConst UnusedArg $ \[a1,o]->(1,elemT a1)], \[a1,o,a2]->("\\a needleFn->let needle = ("++coerceTo (elemT a1) (ret a2)++" needleFn) in \
         \map (\\i->fromIntegral i+1) $ " ++
            if o==OptionNo then "elemIndices needle a"
            else "findIndices (\\e->e /= needle) a"
      ) ~> vList1 VInt),
   extendOp "`=" [reverseRep, filterRep] equivalentOrderReason ([list, Fn ReqArg UnusedArg $ \[a1]->(1,elemT a1)], \[a1,a2]->"\\a f->groupBy (onToBy f) a" ~> vList1 a1),
   extendOp "=~" [reverseRep, mapRep] equivalentOrderReason ([list, AutoOption "nosort", Fn ReqArg UnusedArg $ \[a1,_]->(1,elemT a1)], \[a1,o,_]->"\\a f->\
      \map (map (\\(e,_,_)->e)) $"++
      (if o==OptionYes then "sortOn (\\((_,_,ind):_)->ind) $" else "")++
      "groupBy (onToBy (\\(_,fa,_)->fa)) $\
      \sortOn (\\(_,fa,_)->fa) $\
      \zip3 a (map f a) [0..]" ~> vList1 a1),
   extendOp "or" [reverseRep, mapRep] equivalentOrderReason ([list, AutoOption "tbd", Fn ReqConst UnusedArg $ \[a1,_]->(1,elemT a1)], \[a1,o,a2]->lazyOr [a1] (ret a2)),
   opM("`<", [14], [list, BinCode 13], "sort" ~> a1),
   opM("`'",[14],[list, BinCode 3], \[a1] ->
      case a1 of
         VList [VList _] -> "transpose" ~> [a1]
         VList (_:_:_) -> unzipTuple a1
         otherwise -> "transpose.(:[])" ~> [VList [a1]]
      ),
   extendOp "=\\" [lengthRep, mapRep] (shorterReason"just take length without map") ([list, auto, fn2 (const []), fnx (\[a1,a2]->(length $ ret a2, elemT a1 ++ ret a2))], foldrFn "(scanl . flip)" ~> VList .ret.a2),
   extendOp "=\\" [lengthRep, mapRep] (shorterReason"just take length without map") ([list, Fn ReqArg UnusedArg $ \[a1]->(length $ elemT a1,concat $ replicate 2 $ elemT a1)], foldr1Fn "(scanl1 . flip)" "take 0.(:[])" ~> VList .elemT.a1),
   extendOp "=\\" [lengthRep, mapRep] (shorterReason"just take length without map") ([list, FakeAuto $ italic "tuple", Fn ReqConst UnusedArg $ \[a1]->(length $ elemT a1,concat $ replicate 2 $ elemT a1), fn (\[a1,a2]->elemT a1 ++ ret a2)], foldrFn2 "(scanl . flip)" ~> VList .ret.a2),
   opM("`\\", [14], [list, BinCode 12, FoldMode], \[a1,rt]->"\\a f->f (scanl1.flip,const[]) a" ~> VList (ret rt)),
   opM("`/", [14], [list, BinCode 11, FoldMode], \[a1,rt]->"\\a f->f (foldr1,id) a" ~> ret rt),
   opM("`*", [9,0], [listOf int], \[a1]->let (uzT,uzF)=unzipTuple a1 in
         appFst uzT "product" ++ "." ++ uzF ~> VInt : tail uzT
      ),
   extendOp "`_" [dropRep, mapRep] equivalentOrderReason ([AutoDefault int 2, list],"\\n a->\
      \if n>0 then subsequencesN n a \
      \else if n==0 then subsequences a \
      \else repeatedSubsequencesN (-n) a"~>vList1.a2),
   opM("`*",[9,0],[listOf list],"sequence"~>a1),
   extendOp "``p" [reverseRep, addRep, tildaRep] equivalentOrderReason ([list], "permutationsSaneOrder"~>vList1.a1),
   extendOpHelper ["`-"] (shorterReason"just use -") ("`:", [14], [listToBeReferenced, BinCode 8, sameAsA1], "\\a b->[a,b]" ~> vList1.a1),
   binarySetOp "`-" 8 "a-b",
   binarySetOp "`&" 5 "reverse $ reverse a - (a - b)",
   binarySetOp "`|" 6 "a ++ (b - a)",
   binarySetOp "`^" 7 "(a-b) ++ (b-a)",
   opM("`$",[14],[list, BinCode 4],"nub"~>a1),
   commutativeExtension [12,0]
      ("!=", [AutoDefault num 0,num], "(abs.).(-)" ~> VInt)
      ("`&", [num,AutoDefault num (-2)],".&."~>orChr),
   commutativeExtension [11,0]
      ("`|",[AutoDefault num 1,num],".|."~>orChr)
      ("`^",[num,AutoDefault num 1],"xor"~>xorChr),
   opM("%~", [14], [list, BinCode 0, AutoNot $ fn (elemT.a1)], \[a1,_]->
      "\\a f->let r = chunksOf 2 $ (split.dropFinalBlank.condense.whenElt) f a \
      \in map (\\c->let (a,b)=splitAt 1 c in (concat b,concat a)) r" ~> VList [a1,a1]),
   extendOp "`%" [ifRep, intRep] (shorterReason"don't use an if, just provide either the true or false clause depending on the constant)") ([list,OrAuto "default" any1], \[a1,a2]->
      (if a2==OptionYes then
         "\\a -> splitOn ["++defaultValue (elemT a1)++"] a"
      else
         "\\a needle->splitOn ("++coerceTo [a1] [a2]++" needle) a") ~> vList1 a1),
   opM("-~",[9,0],[str {-could be more general, but probably not useful -}],\[a1]->let cond="(not."++truthy (elemT a1)++")" in
      "dropWhileEnd "++cond++" . dropWhile "++cond~>a1),
   extendOp "`$" [lengthRep, consRep] (shorterReason"just add 1 to the length") ([num], "signum" ~> VInt),
   opM(["`",")"], [12], [char, BinCode 2], "myOrd.toUpper.myChr"~>a1),
   opM(["`","("], [12], [char, BinCode 7], "myOrd.toLower.myChr"~>a1),
   extendOp "ch" [lengthRep, rangeRep] (shorterReason"use max ~ (0) if you wanted that") ([AutoDefault int 256], "id" ~> xorChr.(VChr:)),
   extendOp "tbd" [lengthRep, rangeRep] (shorterReason"use max ~ (0) if you wanted that") ([autoTodo char], undefinedImpl),
   extendOp "`p" [ifRep, intRep] (shorterReason"don't use an if, just provide either the true or false clause depending on the constant)") ([num], inspect.a1 ~> vstr),
   extendOp "``@" [lengthRep, consRep, tildaRep] (shorterReason"just add 1 to the length") ([num], "toBase 2" ~> vList1 VInt),
   extendOp "hex" [(takeRep),('<',0),(strRep)] (shorterReason"init of a constant string can just be hardcoded") ([any1], \[a1]->case a1 of
      VInt -> "\\i -> sToA $ (if i < 0 then \"-\" else []) ++ showHex (abs i) []" ~> vstr
      VList [VChr] -> "parseNum 16" ~> VInt
      VChr -> "parseNum 16 . (:[])" ~> VInt
      VList [VInt] -> "fromBase 16" ~> VInt
      otherwise -> "hex is undefined for " ++ show a1 ~> VInt
      ),
   extendOp "`D" [addRep,tildaRep,tildaRep] (shorterReason"+~~ = 3") ([ParseArg "int" staticIntParser], ((\[StaticInt v1]->do
      modify $ \s -> s { pdDataUsed = True }
   extendOp "`@" [sumRep, consRep] ("just use a+b") ([AutoDefault int 10, autoTodo num], "toBase"~>vList1 .a1),
   opM("`@",[11,0], [AutoDefault num 10,list {-todo 1d-}], "\\base a->fromBase (abs base) (if base < 0 then map (\\c->fromIntegral $ fromMaybe (fromIntegral c) $ elemIndex c printables) a else a)"~>a1),
   extendOp "`." [reverseRep, strRep] (shorterReason"just write the string in reverse") ([AnyS, AutoOption "inf", fnx (\[a1,o1]->(length $ ret a1, ret a1))],
      \[a1,o1,a2]->"\\i f->"++(if o1==OptionYes then "iterate" else "iterateWhileUniq") ++"("++coerceTo (ret a1) (ret a2)++".f) (i())" ~> VList (ret a1)),
   extendOp ".~~" [subtractRep,tildaRep,tildaRep] (shorterReason"-~~ = 0") ([AnyS, fn $ \[a1]->[fst $ promoteList (ret a1)]], \[a1,a2]->
      let (a1T,a1C) = promoteList (ret a1) in
         "\\a f->appendUntilNull ("++a1C++"(a())) ("++coerceTo [a1T] (ret a2)++".f)" ~> a1T),
   opM(";~",[6,0], [AnyS, Fn ReqArg UnusedArg $ \[a1]->(1,ret a1)],
      (\[a1,a2]->
         let a1L = length $ ret a1
             a2L = length $ ret a2 in
         "\\x f->"++flattenTuples a2L 1 ++ "(f $ x(),f)" ~>
         ret a2 ++ [VFn (ret a1) (ret a2)]
         )),
   opM("==",[6,0],[AnyS, Fn ReqConst UnusedArg $ \[a1]->(1,ret a1)],\[a1,a2]->
      let (coercedType, coerceFnA, coerceFnB) = coerce (ret a1) (ret a2) in
         "\\a b->bToI$"++coerceFnA++"(a())=="++coerceFnB++"b" ~> VInt),
   extendOp "``;" [multRep, tildaRep, tildaRep] (shorterReason"*~~ = -2") recursionDef,
   extendOp "`;" [setRep,setRep] (shorterReason"multiple assignments on the same thing is useless, just use 1") recursionDef,
   opM("`#", [15,0], [AutoOption "nosalt", any1, ParseArg "int" intParser], (\[o,a1,a2]->do
      let salt = o/=OptionYes
      modify $ \s -> s { pdDataUsed = pdDataUsed s || salt }
   opM ("fsb", [], [list, int, AutoDefault int (2^128), AutoNot $ fn (\_->[VList [VInt]])], \[a1,_,_,_]->"\\inputs modn stopat goalChecker->toHex $\
      \fromMaybe (-1) $ find (\\salt->goalChecker (map (\\input->(fromIntegral$hlist$("++flatten(head $ elemT a1)++")input++toBase 256 salt) `mod` modn) inputs)) [0..stopat] "~>vstr),
   opM ("fs", [], [list, int, AutoDefault int (2^128), list], \[a1,_,_,a4]->"\\inputs modn stopat goal->toHex $\
      \fromMaybe (-1) $ find (\\salt->and $ zipWith ("++(if cidim [a4]>1 then "elem" else "(==)")++") (map (\\input->(fromIntegral$hlist$("++flatten(head $ elemT a1)++")input++toBase 256 salt) `mod` modn) inputs) goal) [0..stopat] "~>vstr),
   opM("p", [], [any1], inspect.a1 ~> vstr),
   extendOp "error" [diffRep, rangeRep, tildaRep, strRep] "- ,~ str is invalidly typed" ([any1], \[a1]->"\\s->errorWithoutStackTrace $ \"Error: \"++aToS ("++(if a1==vstr then "" else inspect a1)++" s)++\" (\"++"++ show version++"++\")\"" ~> a1),
   opM("testCoerce2", [], [any1, any1], testCoerce2 ~> vstr),
   opM("testCoerceToInt", [], [any1], testCoerceTo [VInt]),
   opM("testCoerceToChr", [], [any1], testCoerceTo [VChr]),
   opM("testCoerceToListInt", [], [any1], testCoerceTo [VList [VInt]]),
   opM("testCoerceToStr", [], [any1], testCoerceTo [vstr]),
   opM("testCoerceToListListInt", [], [any1], testCoerceTo [VList [VList [VInt]]]),
   opM("testCoerceToListStr", [], [any1], testCoerceTo [VList [vstr]]),
   opM("testCoerceToX", [], [AnyS, AnyS], \ts -> "\\a b->" ++ snd (testCoerceTo (ret $ a1 ts) (ret $ a2 ts)) ++ "(b())" ~> (ret $ a1 ts)),
   opM("testFinish", [], [any1], fst . finish . a1 ~> vstr),
foldr1Fn foldOp initOp = (\[a1,a2]->"\\a f->if null a then "++initOp++"$"++defaultValue (elemT a1)++" else "++foldOp++" (\\x y->"++coerceTo (elemT a1) (ret a2)++"$f$"++flattenTuples(length $ elemT a1)(length $ elemT a1)++"(x,y)) a")
foldrFn foldOp = (\[a1,a2,a3]->"\\a i f->"++foldOp++" (\\x y->"++coerceTo (ret a2) (ret a3)++"$f$"++flattenTuples(length $ elemT a1)(length $ ret a2)++"(x,y)) (i()) a")
foldrFn2 foldOp = (\[a1,a2,a3]->"\\a i f->"++foldOp++" (\\x y->"++coerceTo (ret a2) (ret a3)++"$f$"++flattenTuples(length $ elemT a1)(length $ ret a2)++"(x,y)) (i) a")
mapFn = (\[a1,a2]->"(\\a f->map f a)")
recursionDef = ([AnyS, fnx (\[a1]->(0, ret a1++[InvalidType]))],
      "\\x f -> memoFix (\\rec x->let (a,b,c)=f ("++flattenTuples (length $ ret a1) 1++"(x,rec)) in if a then c else b) (x())" ~> ret (head $ tail $ ret a2))
ifBodyArgs =
   [ Fn ReqDontCare OptionalArg $ \a1 -> (1,a1)
   , OrAuto "default" $ fnx (\[a1,a2]->(length$ret a2,[]))]
ifImpl [a1,a2,a3] =
   then "\\c a->if "++truthy [a1]++" c then a c else "++defaultValue (ret a2) ~> ret a2
   else let (coercedType, coerceFn) = coerceEither (ret a2) (ret a3)
      in "\\c a b->"++coerceFn ++ "$ (iff."++truthy [a1]++") c (a c) (b())" ~> coercedType
zipImpl [a1,a2] = "zipWith (\\a b->"++flattenTuples (length$elemT a1) (length$elemT a2) ++ "(a,b))" ~> VList (elemT a1++elemT a2)
binarySetOp name binrep code =
   opM(name, [14], [list, BinCode binrep, AutoOption "uniq",OptionalFn (\[a1,_]->(1,elemT a1)),any1], \[a1,o1,optionalFn,a2]-> let
      (coercedType, coerceFnA, coerceFnB) = coerce [a1] bt
      replaceMinus = concatMap (\c->if c=='-' then "`minus`" else [c])
      maybeNub = if o1==OptionYes then "(nubBy (onToBy snd))" else "id"
      (byFn,bcode,bt) = case optionalFn of
         otherwise -> ("ff","bb",[a2])
      zipBy = "(\\x -> zip x $ map " ++ byFn ++ " x)"
    in
      "\\aa ff bb->let minus=deleteFirstsBy (onToBy snd);\
      \a="++maybeNub++"$"++zipBy++"$"++coerceFnA++"aa;\
      \b="++maybeNub++"$"++zipBy++"$"++coerceFnB++bcode++" in \
      \map fst $ "++maybeNub++"$"++replaceMinus code~>coercedType)
addHigherValueDeBruijnOps ops = concat (zipWith (\(nib,lit) n->
      opM(lit, nib, [], argn n))
   deBruijnArgReps [1..])
      ++ map convertNullNib ops
unvecOrChr [a1,a2] | a1==a2 = a1
unvecOrChr [a1,a2] | a1/=a2 = error "unhandled, should be impossible 28"
unvecOrChr a = orChr a
binOp '+' a = ([xorChr a], "+")
binOp '*' a = ([VInt], "*")
binOp '-' a = ([xorChr a], "-")
binOp '/' a = ([VInt], "safeDiv")
binOp '%' a = ([VInt], "safeMod")
binOp '^' a = ([VInt], "^")
binOp ']' a = ([unvecOrChr a], "max")
binOp '[' a = ([unvecOrChr a], "min")
binOp '!' a = ([VInt], "(abs.).(-)")
binOp ':' [a,b] = let
      (ap,apFn) = promoteList (ret a)
      (coercedType, coerceFnA, coerceFnB) = coerce [ap] (ret b)
   in
      "\\a b->("++coerceFnA++"$"++apFn++"(a()))++"++coerceFnB++"(b())"~>coercedType
binOp '|' [a1,a2] = lazyOr [a1] [a2]
binOp '&' [a1,a2] = lazyAnd [a1] [a2]
binOp '=' [a1,a2] = (elemT a1, "\\a i->if null a then "++defaultValue (elemT a1)++" else lazyAtMod a (fromIntegral i - 1)")
binOp '?' [a1,a2] = ([VInt], "\\a e->fromIntegral$1+(fromMaybe (-1) $ elemIndex e a)")
allOps = sortOn opSpecificity $ addHigherValueDeBruijnOps $ concat rawOps
simpleOps = sortOn opSpecificity $ addHigherValueDeBruijnOps $ filter isOpSimple $ map (\l ->
   if length l == 6 then l!!3 else last l)
      rawOps
typeToStr (Cond desc _) = Just desc
typeToStr Auto = Just "~"
typeToStr (FakeAuto _) = Nothing
typeToStr (BinCode b) = Nothing
typeToStr (LitCode b) = Nothing
typeToStr (AutoOption _) = Nothing
typeToStr (AutoNot s) = typeToStr s
typeToStr (OrAuto _ a) = typeToStr a
typeToStr (Fn ReqDontCare _ _) = Just "fn"
typeToStr (Fn ReqArg _ _) = Just "reqfn"
typeToStr (Fn ReqConst _ _) = Just "const"
typeToStr (AutoDefault t _) = typeToStr t
typeToStr (AutoData t) = typeToStr t
typeToStr (ParseArg desc _) = Just $ "{"++desc++"}"
typeToStr (OptionalFn _) = Just $ "fn?"
typeToStr (ZipMode) = Just "zipop"
typeToStr (FoldMode) = Just "foldop"
typeToStr (CharClassMode) = Just "chclass"
typeToStr AnyS = Just "any*"
typeToStr _ = Just $ "unknown"
makeOp priority (lit, nib, t, behavior) = [(priority, toLitSpec lit, nib, (t, behavior))]
opM (lit,nib,args,impl) = makeOp True (lit,nib,args,CodeGen $ toImpl impl)
op ((lit,nib),args,impl) = opM(repToLit lit,[nib],args,impl)
lowPriorityOp (lit,nib,args,impl) = makeOp False (lit,nib,args,CodeGen $ toImpl impl)
shorterReason how = "There is a shorter way to achieve this effect, " ++ how ++ "."
equivalentOrderReason = "Use the other equivalent operation order."
associativeReason = "Use the other operation order for this associative op to accomplish this. E.g. (a+b)+c instead of a+(b+c)."
commutativeReason = "Use the other operator order for this commutative op to accomplish this. E.g. (b+a) instead of (a+b)."
litExtError invalidLit lit reason = LitWarn $ "You used an op combo that has been remapped to an extension in the binary form.\nYou wrote: " ++ formatInvalidLit invalidLit ++ " but this actually will mean: " ++ lit ++ "\n" ++ reason ++ " For more infromation see "++webpage++"/tutorial_ancillary.html#extensions or if you are learning try \"nibbles -simple\" to disable all extensions." where
   formatInvalidLit = concatMap $ \l -> if l==litDigit then "[digit]" else l
makeExtendOp priority invalidLit reason (lit, nib, t, impl) =
   makeOp priority (invalidLit, [], map toLitCode t, litExtError invalidLit lit reason)
      ++ opM (lit, nib, map toBinCode t, impl)
makeExtendOpM priority invalidLit reason (lit, nib, t, impl) =
   makeOp priority (invalidLit, [], map toLitCode t, litExtError invalidLit (concat lit) reason)
      ++ opM (lit, nib, map toBinCode t, impl)
extendOpHelper = makeExtendOp True
extendOpHelperM = makeExtendOpM True
lowPriorityExtendOp = makeExtendOp False
extendOp name from reason impl = extendOpHelper (map (repToLit.fst) from) reason (name, (map snd from), fst impl, snd impl)
extendOpM name from reason impl = extendOpHelperM (map (repToLit.fst) from) reason (name, (map snd from), fst impl, snd impl)
repToLit r = if r == '\0' then [] else [r]
commutativeExtension bin (rep1,t1,impl1) (rep2,t2,impl2) = concat [
   extendOpHelper [rep2] commutativeReason (rep1, bin, init t1 ++ [andC (last t1) nArgLarger], impl1),
   extendOpHelper [rep1] commutativeReason (rep2, bin, t2, impl2),
   extendOpHelper [rep2] commutativeReason (rep1, [], t1, impl1)]
undefinedImpl = (VInt,"asdf")
toUntypedImpl hs = noArgsUsed { implCode=hsParen $ hsAtom hs }
toUntypedImpl2 (ts,hs) = (ts,toUntypedImpl hs)
class OpImpl impl where
instance OpImpl ([VT] -> ParseState ([VT], Impl)) where toImpl = id
instance OpImpl ([VT] -> VT, [VT] -> String) where
   toImpl (f1,f2) context = return ([f1 context], toUntypedImpl $ f2 context)
instance OpImpl ([VT] -> VT, String) where
   toImpl (f1,s) context = return ([f1 context], toUntypedImpl s)
instance OpImpl (VT, [VT] -> String) where
   toImpl (t,f2) context = return ([t], toUntypedImpl $ f2 context)
instance OpImpl (VT, String) where
   toImpl (t,s) context = return ([t], toUntypedImpl s)
instance OpImpl ([VT] -> (VT, String)) where
   toImpl f context = return ([t],toUntypedImpl s) where (t, s) = f context
instance OpImpl ([VT] -> [VT], [VT] -> String) where
   toImpl (f1,f2) context = return (f1 context, toUntypedImpl $ f2 context)
instance OpImpl ([VT] -> [VT], String) where
   toImpl (f1,s) context = return (f1 context, toUntypedImpl s)
instance OpImpl ([VT], [VT] -> String) where
   toImpl (t,f2) context = return (t, toUntypedImpl $ f2 context)
instance OpImpl ([VT], String) where
   toImpl (t,s) context = return (t, toUntypedImpl s)
instance OpImpl ([VT] -> (OptionalLets, String)) where
   toImpl f context = do
      let (OptionalLets t,impl)=toUntypedImpl2 $ f context
      return (t,impl { implUsed = OptionalArg } )
instance OpImpl ([VT] -> ([VT], String)) where
   toImpl f context = return $ toUntypedImpl2 $ f context
instance OpImpl ([VT] -> ParseState (VT,String)) where
   toImpl implMonad ts = do
      (t,hs) <- implMonad ts
      return ([t], toUntypedImpl hs)
instance OpImpl ([VT] -> ParseState ([VT],String)) where
   toImpl implMonad ts = do
      (t,hs) <- implMonad ts
      return (t, toUntypedImpl hs)
instance OpImpl () where toImpl _ ts = return $ (ts, toUntypedImpl "id")
instance OpImpl (ParseState Impl) where
   toImpl implMonad ts = do
      impl <- implMonad
      return ([implType impl], impl)
instance ToLitSpec [String] where toLitSpec s = s
instance ToLitSpec String where toLitSpec s = [s]
infixr 1 ~>
a~>b = (b,a)
convertNullNib (isPriority, lit, nib, op) = (isPriority, lit, if null nib
      then [16, error $ "attempt to convert "++(concat lit)++" to bin (it is only for literate mode)"]
      else nib
   , op)
vList1 x = VList [x]
dup a = [a,a]
byType ft = ft . last . mtdTypes
exactType t = byType (==t)
int = Cond "int" $ exactType VInt
specialInt = Cond "spec" $ \mtd -> error $ case (last $ mtdTypes mtd) of
   (VList _) -> "here"
   otherwise -> "there"
char = Cond "chr" $ exactType VChr
str = Cond "str" $ exactType vstr
fn f = fnx $ \ts->(1, f ts)
fn2 f = fnx $ \ts->(2, f ts)
fnx = Fn ReqDontCare UnusedArg
num = Cond "num" $ byType isNum
vec = Cond "vec" $ byType $ const True
list = Cond "[*]" $ byType isList
listToBeReferenced = Cond "[a]" $ byType isList
any1 = Cond "any" $ byType $ const True
auto = Auto
listOf (Cond desc t) = Cond ("["++desc++"]") $ \mtd -> let lastArg = last $ mtdTypes mtd in
   isList lastArg && t (mtd { mtdTypes=reverse $ elemT $ lastArg })
nonTupleList = Cond "[1]" $ byType (\t -> isList t && length (elemT t) == 1)
elemOfA1 = Cond "a" $ \mtd -> let [a1,a2] = mtdTypes mtd in
   isList a1 && head (elemT a1) == a2
sameAsA1 = Cond "[a]" $ \mtd -> let [a1,a2] = mtdTypes mtd in
   a1 == a2
nArgLarger = Cond ">" $ \mtd -> let (op1b:oprb) = mtdNibs mtd
   in last oprb > op1b
andC (Cond as af) (Cond bs bf) = Cond (bs++as) (\mtd -> af mtd && bf mtd)
orC (Cond as af) (Cond bs bf) = Cond (as++"|"++bs) (\mtd -> af mtd || bf mtd)
testCoerce2 [a1,a2] = "const $ const $ sToA $ " ++ show (if ct1 == ct2
   then show ct1
   else "flipped mismatch: " ++ show ct1 ++ "," ++ show ct2)
   where
      ct1 = coerce2[a1][a2]
      ct2 = coerce2[a2][a1]
testCoerceTo to a1 =  (to, coerceTo to a1)
isExtension isExtOptFn (lit, nib, op) = (isPrefixOf [16] nib) || (length nib > 1 && length lit > 1 || any isExtOptFn (fst op) || elem '~' lit )
isExtOpt t = case t of
   Auto -> True
   Cond desc _ -> elem '>' desc
   BinCode _ -> True
   Fn ReqConst _ _ -> True
   otherwise -> False
isOpSimple (isPriority, lits, nib, op@(types,_)) =
   (not (isExtension isExtOpt (lit, nib, op) || null nib)
   ||elem lit whitelist)
      && not (elem lit blacklist)
      && not (any isSpecialMode types)
   where
      lit = concat lits
      whitelist = ["ct","p"]
      blacklist = ["tbd"]
      isSpecialMode CharClassMode = True
      isSpecialMode (BinCode _) = True
      isSpecialMode ZipMode = True
      isSpecialMode _ = False
opSpecificity (_,lit,bin,(args, _))  = let replen = if head bin==16 then length (concat lit) else length bin in -(replen + sum (map argSpecificity args))
argSpecificity (BinCode _) = 1
argSpecificity Auto = 1
argSpecificity _ = 0
autoTodoValue = -88
autoTodo t = AutoDefault t autoTodoValue
toLitCode (BinCodeRep (l,b)) = LitCode l
toLitCode a = a
toBinCode (BinCodeRep (l,b)) = BinCode b
toBinCode a = a
italic s = "<i>"++s++"</i>"
   staticIntParser,
   intParser,
   strParser,
   chrParser,
   parseDataExpr,
   parseCountTuple,
   cp,
   tildaOp,
   onlyCheckMatch,
   match,
   parseError,
   parseLitWarning,
   nextOffset,
   consumeWhitespace,
   litDigit,
   empty,
   fromByte,
   toByte,
   parse1Nibble,
   errorUnderLine,
   onlyCheckMatchIdentifier,
   isIdentifierChar,
   getLitParseCoordinates) where
toByte (a:b:_)=chr $ 16 * a + b
fromByte b=[ord b `div` 16, ord b `mod` 16]
sLit f s cp = consumeWhitespace $ Lit f s cp
litDigit = ""
tildaOp = ([0],["~"])
parseInt (Nib (0:s) cp) = case parseNibInt s 0 (cp+1) of
   (6,rest) -> (-10,rest)
   (9,rest) -> (-7,rest)
   (n,rest) -> (-1-n,rest)
parseInt (Nib s cp) = case parseNibInt s 0 cp of
   (7,rest) -> (10,rest)
   (10,rest) -> (7,rest)
   (n,rest) -> (n,rest)
parseInt (Lit f ('0':s) cp) = (0, sLit f s (cp+1))
parseInt (Lit f ('-':s) cp) = let (n, rest) = parseInt $ Lit f s (cp+1) in
   (-n, rest)
parseInt (Lit f s cp) = case readP_to_S (gather readDecP) s of
   [((used, n), rest)] -> (n, sLit f rest (cp+length used))
   _ -> error $ "unparsable int: " ++ s
parseInt16 (Lit f s cp) = case readP_to_S (gather readHexP) s of
   [((used, n), rest)] -> (n, sLit f rest (cp+length used))
   _ -> error $ "unparsable int: " ++ s
parseNibInt(f:s) a cp
   | f>=8 = (c,Nib s (cp+1))
   | otherwise = parseNibInt s c (cp+1)
   where c=a*8+toInteger f`mod`8
intToNib (-7) = [0,1,9]
intToNib (-10) = [0,14]
intToNib n | n<0 = 0:intToNibH (-1 - n)
intToNib n = intToNibH n
intToNibH n = init digits ++ [last digits + 8]
   where digits = map digitToInt $ showOct n ""
parseStr(Nib (a:s) cp)
   | a8==0 = cont '\n' 1
   | a8==1 = cont ' ' 1
   | c==32 = if a<8
      then ([""], Nib (tail s) (cp+2))
      else let (r2,code2) = parseStr $ Nib (drop 1 s) (cp+2)
         in ("":r2,code2)
   | c==127 = cont (toByte $ tail s) 4
   | otherwise = cont (chr c) 2
   where
      a8=a`mod`8
      c=a8*16+if null s
         then error "unterminated binary string with partial character"
         else head s
      cont ch used = if a>=8
         then ([[ch]], after)
         else let (rest:rest2, rs)=parseStr after in ((ch:rest):rest2, rs)
         where after = Nib (drop (used-1) s) (cp+used)
parseStr (Lit f s cp) =
   case readP_to_S (gather Lex.lex) s of
   [((used, Lex.String str), rest)] ->
      let after = sLit f rest (cp+length used) in
      case rest of
         ('"':_) ->
            let (strRest, after2) = parseStr after
            in (str:strRest, after2)
         _ -> ([str], after)
   _ -> error $ "unparsable string: " ++ s
strToNib [""] = [2,0]
strToNib [s] = (concatMap (\(c,last)->let oc = ord c in case c of
   '\n' -> [last]
   ' ' -> [1+last]
   c | oc > 126 || oc < 32 -> [last+7, 15, div oc 16, mod oc 16]
   _ -> [last+div oc 16, mod oc 16]
   ) (zip s $ take (length s - 1) (repeat 0) ++ [8]))
strToNib ("":rest) = [10,0] ++ strToNib rest
strToNib (s:rest) =
   let nib1 = strToNib [s++" "]
   in init nib1 ++ [10,0] ++ strToNib rest
specialChars = "\n /.,`a@A0"
specialChars127n = '\127':tail specialChars
swap127 '\n' = '\127'
swap127 '\127' = '\n'
swap127 c = c
parseChr(Nib [] cp) = error "unterminated char"
parseChr(Nib (c:rest) cp)
   | c < 2 = (specialChars !! c, Nib rest (cp+1))
   | c > 7 = (specialChars !! (c-6), Nib rest (cp+1))
parseChr(Nib [_] cp) = error "unterminated char"
parseChr(Nib (a:b:rest) cp) =
   case elemIndex c specialChars127n of
   Just i -> case uncons rest of
      Nothing -> error "unterminated binary char"
      Just (b2,rest2) -> (swap127 $ toByte [if i < 2 then i else i+6,b2], Nib rest2 (cp+3))
   Nothing -> (c, Nib rest (cp+2))
   where c = toByte [a,b]
parseChr (Lit f s cp) = case readP_to_S (gather Lex.lex) s of
   [((used, Lex.Char char), rest)] -> (char, sLit f rest (cp+length used))
   _ -> error $ "unparsable char: " ++ s
chrToNib c =
   case elemIndex c specialChars of
   Just i -> [if i < 2 then i else i + 6]
   Nothing -> if c >= chr 32 && c < chr 127
      then fromByte c
      else let [a,b] = fromByte (swap127 c)
         in fromByte (specialChars127n!!(if a<2 then a else a-6))++[b]
parseData l@(Lit _ _ _) = if empty rest then n else error $ "program must be empty after storing ~ integer data. See "++webpage++"/tutorial_minutiae.html#data " where
   (n,rest)=parseInt16 l
parseData (Nib b _) = fromBase 16 (map fromIntegral $ padSafeDat b)
parseCountTuple = do
   matched <- match tildaOp
   if matched then fmap (+1) parseCountTuple else return 0
cp (Lit _ _ cp) = cp
cp (Nib _ cp) = cp
empty (Nib [op] _) | op == uselessOp = True
empty (Nib [] _) = True
empty (Lit _ [] _) = True
empty _ = False
nextOffset (Nib (c:s) cp) = Nib s (cp+1)
nextOffset (Lit f (c:s) cp) = Lit f s (cp+1)
onlyCheckMatch (Nib s cp) (needle, _) = if isPrefixOf needle s
   then Just $ Nib (drop (length needle) s) (cp+length needle)
   else Nothing
onlyCheckMatch lit@(Lit _ _ _) (_, []) = Just lit
onlyCheckMatch lit@(Lit _ _ _) (_, needle:s) = match1Lit lit needle >>= flip onlyCheckMatch (undefined, s)
match1Lit lit@(Lit f s cp) needle
   | null s = Nothing
   | needle == litDigit =
      case s of
         (c:_) | isDigit c -> Just lit
         ('-':c:_) | isDigit c -> Just lit
         otherwise -> Nothing
   | needle == "\"" && '"' == head s = Just lit
   | needle == "\'" && '\'' == head s = Just lit
   | isPrefixOf needle s = Just $ sLit f (drop (length needle) s) (cp+length needle)
   | otherwise = Nothing
match (nib,lit) = do
   code <- gets pdCode
   case onlyCheckMatch code (nib,lit) of
      Just nextCode -> do
         modify $ \s -> s { pdCode = nextCode }
         appendRep (nib,concat lit)
         return True
      Nothing -> return False
onlyCheckMatchIdentifier rawCode@(Lit f code cp) =
   if not (empty rawCode)
      && isIdentifierChar (head code)
      && not (isDigit $ head code) then
         let (name,rest) = span isIdentifierChar code in
            Just $ (name, sLit f rest (cp+length name))
   else Nothing
onlyCheckMatchIdentifier _ = Nothing
isIdentifierChar c | isAlphaNum c = True
isIdentifierChar '_' = True
isIdentifierChar c = False
generateErrorMsg msg = do
   s <- gets pdCode
   return $ msg ++ "\n" ++ case s of
      Lit f s cp -> literateError f cp
      Nib s cp -> "at nibble #" ++ show cp ++ "\nnext nibble is" ++ fromMaybe "" (at s 0>>=Just . show)
parseLitWarning msg = generateErrorMsg msg >>= addLitWarning
parseError msg = do
   generateErrorMsg msg >>= errorWithoutStackTrace
errorUnderLine line charno = let arrows = replicate (charno-1) ' ' in
   "\n" ++ arrows ++ "v"
   ++ "\n" ++ line
   ++ "\n" ++ arrows ++ "^"
literateError s cp =
   "at line: " ++ show lineno
      ++ ", char: " ++ show charno
      ++ errorUnderLine line charno
   where
      (lineno,charno) = getLitParseCoordinates s cp
      line = lines s !! (lineno-1)
getLitParseCoordinates s cp = (lineno,charno)
   where
      prev = take (cp+1) s
      lineno = length $ lines prev
      line = lines s !! (lineno-1)
      charno = 1+(fromMaybe (cp+1) $ elemIndex '\n' $ reverse prev) - 1
consumeWhitespace n@(Nib _ _) = n
consumeWhitespace l@(Lit _ [] _) = l
consumeWhitespace (Lit f (c:s) cp)
   | c=='#' = sLit f rest (cp+1+length comment)
   | isSpace c = sLit f s (cp+1)
   | otherwise = Lit f (c:s) cp
      where (comment, rest) = break (=='\n') s
intParser = do
   ~(_, hs) <- staticIntParser
   return (VInt, hs)
staticIntParser = do
   ~(n, rest) <- gets $ parseInt . pdCode
   modify $ \st -> st { pdCode=rest }
   return (StaticInt n, flatHs $ i n)
strParser = do
   ~(ss, rest) <- gets $ parseStr . pdCode
   modify $ \st -> st { pdCode=rest }
   let (et,val) = case ss of
        [s] -> (VChr, "sToA " ++ show s)
        _ -> (vstr, "map sToA " ++ show ss)
   return (VList [et], val)
chrParser = do
   ~(s, rest) <- gets $ parseChr . pdCode
   modify $ \st -> st { pdCode=rest }
   appendRep (chrToNib s,tail $ show s)
   return (VChr, "myOrd " ++ show s)
parseDataExpr = do
   dat <- gets $ parseData . pdCode
   let datNibs = map fromIntegral (toBase 16 dat)
   progNibs <- gets pdNib
   appendRep (padSafeDat datNibs," "++showHex dat "")
   return dat
padSafeDat = reverse . map (\e ->
   if e == uselessOp then 0
   else if e == 0 then uselessOp
   else e)
parse1Nibble name [] = parseError $ "symbol not found in " ++ name
parse1Nibble name ((nib,(lit,impl)):rest) = do
   match <- match ([nib],[[lit]])
   if match then impl else parse1Nibble name rest
data RBracket = EofBracket | ListBracket | TupleBracket
rBracketStr EofBracket = "EOF"
rBracketStr ListBracket = "]"
rBracketStr TupleBracket = ")"
isNibblesArg ('(':_) = True
isNibblesArg ('[':_) = True
isNibblesArg ('"':_) = True
isNibblesArg ('\'':_) = True
isNibblesArg ('-':s) = isPositiveIntegerArg s
isNibblesArg s = isPositiveIntegerArg s
isPositiveIntegerArg (c:s) = isDigit c
   && not (isSuffixOf ".nbl" s)
   && not (isSuffixOf ".nbb" s)
getArgTypes s = case getArgTypesH [EofBracket] s of
   (t, "") -> t
   (_, s) -> argError s "expecting eof"
 where
   argError rest msg = errorWithoutStackTrace $ "arg io error: "++msg ++ errorUnderLine s (length s - length rest)
   appendFst x (fst, rest) = (x:fst, rest)
   catFst x (fst, rest) = (x++fst, rest)
   lstrip s = fromMaybe s $ stripPrefix " " s
   assert1 s [] = argError s "need at least one value"
   assert1 _ [t] = t
   assert1 s _ = argError s "no nested tupes, use flat tuples"
   getArgTypesH rbrackets a@(c:s)
      | isSpace c = getArgTypesH rbrackets s
      ([VInt], rest)
      ([vstr], rest)
      ([VChr], rest)
   getArgTypesH rbrackets a@('[':s) =
      let (et, rest) = parseMultiple (ListBracket:rbrackets) s True in case nub et of
         [] -> argError s $ "lists cannot be empty todo deduce type if needed or assume ints"
         [uniqEt] -> ([VList uniqEt], rest)
         otherwise -> argError s $ "lists types must be heterogeneous, did you mean to use a tuple?"
   getArgTypesH rbrackets a@('(':s) =
      let (ts,rest) = parseMultiple (TupleBracket:rbrackets) s False in
      (map head ts, rest)
   getArgTypesH _ s = argError s $ "cannot parse args (need - | [0-9] | [ | \" | ' ) to start a value\nat: " ++ s
   parseMultiple rbrackets s allowMult = parseMultipleH rbrackets [] (',':s) allowMult
   parseMultipleH rbrackets typesSoFar (',':s) allowMult =
      let (et, rest) = getArgTypesH rbrackets s in
         parseMultipleH rbrackets (typesSoFar++[if allowMult then et else [assert1 s et]]) rest allowMult
   parseMultipleH [EofBracket] typesSoFar "" allowMult = (typesSoFar, "")
   parseMultipleH (ListBracket:_) typesSoFar (']':rest) allowMult = (typesSoFar, rest)
   parseMultipleH (TupleBracket:_) typesSoFar (')':rest) allowMult = (typesSoFar, rest)
   parseMultipleH rbrackets _ rest _ = argError rest $ "mismatched brackets, expecting \"" ++ rBracketStr (head rbrackets) ++ "\""
convertToNibblesType [VChr] = "myOrd"
convertToNibblesType [VList et] = "(map "++convertToNibblesType et++")"
convertToNibblesType [t] = "id"
convertToNibblesType ts = tupleLambda (length ts) $ \varNames -> toTuple $
   zipWith (\var t->convertToNibblesType [t]++" "++var) varNames ts
toLetArgs args = let
   argTypes = map getArgTypes args
   argNames = zipWith (\ts i->
      map (\j->"carg"++show i++"t"++show j) [1..length ts]
      ) argTypes [1..]
   argNamez = map toTuple argNames
   readers = zipWith (\t i->"("++toParser t++")(args!!"++show i++")") argTypes [0..]
   reader = "when (length args /= "++show (length args)++") $ errorWithoutStackTrace $ \"Error: "++show (length args)++" args found at compile time, but \"++show (length args)++\" args found at runtime (pass them in at compile time too!)\";\
      \let "++toTuple argNamez++"="++toTuple readers
   cargs = concat $ zipWith toCargs argNames argTypes
   in (cargs, reader)
toCargs argNames argTypes =
   zipWith (\n t->
         ("z"++n, t, convertToNibblesType [t] ++ " " ++ n)
      ) argNames argTypes
   truthy,
   inspect,
   finish,
   join,
   vectorize,
   coerce,
   coerceEither,
   curryN,
   flattenTuples,
   promoteList,
   promoteListRepeat,
   appTuple,
   firstOf,
   flatten,
   appFst,
   unzipTuple,
   rotateTuple,
   fillAccums,
   defaultValue,
   fullVectorize,
   baseElem,
   cidim,
   tupleLambda,
   minForType,
   maxForType,
   lazyOr,lazyAnd) where
truthy [VInt] = "(>0)"
truthy [VChr] = "(\\c->c>0 && not (isSpace$myChr c))"
truthy [VList _] = "(not.null)"
truthy (xs@(x:_)) = "("++truthy [x]++"."++firstOf xs++")"
minForType ts | length ts > 1 = "("++intercalate "," (map (minForType.(:[])) ts) ++ ")"
minForType [VList _] = "[]"
minForType _ = "(-2^128)"
maxForType ts | length ts > 1 = "("++intercalate "," (map (maxForType.(:[])) ts) ++ ")"
maxForType [VList a] = "(repeat " ++ maxForType a ++ ")"
maxForType _ = "(2^128)"
inspect VInt = "(sToA.show.confirmInt)"
inspect VChr = "(sToA.show.myChr.confirmInt)"
inspect (VList [VChr]) = "(sToA.show.aToS.confirmList)"
inspect (VList et) = "(\\v -> (sToA \"[\") ++ (intercalate (sToA \",\") (map "++inspectElem et++" v)) ++ (sToA \"]\"))"
inspect a = error $ "unhandled type in inspect: " ++ show a
inspectElem [et] = inspect et
inspectElem ts = tupleLambda (length ts) $ \varNames -> "sToA \"(\"++"++
   (intercalate "++sToA \",\"++" $ zipWith (\t v->inspect t ++ v) ts varNames)
   ++"++sToA \")\""
firstOf xs = tupleLambda (length xs) $ head
varNamesN n = map (\tn -> "a"++show tn) [1..n]
curryN n = "(\\f "++intercalate " " varNames++"->f "++toTuple varNames++")"
   where varNames = varNamesN n
tupleLambda n f = "(\\"++toTuple varNames++"->"++f varNames++")"
   where varNames = varNamesN n
appFst t op = tupleLambda (length t) $ \args -> toTuple ((op++" "++head args) : tail args)
unzipTuple (VList ts) = (map (VList.(:[])) ts, "(\\a->"++(toTuple $ map (\i->
   "map "++tupleLambda (length ts) (\args -> args!!i)++"a"
   ) [0..length ts-1]) ++ ")")
rotateTuple n = tupleLambda n (toTuple.rotate) where rotate (a:as) = as++[a]
flattenTuples t1 t2 = "(\\(~"++varsFrom 1 t1++", ~"++varsFrom (1+t1) (t1+t2)++")->("++varsFrom 1 (t1+t2)++"))"
   where varsFrom a b = toTuple $ map (\tn->"a"++show tn) [a..b]
finish t@(VList tt) = (finishH t, sdim tt >= 1)
finish t = (finishH t, False)
finishH (VList tt)
   | d >= 3 = joinC "[]"
   | d == 2 = joinC "firstSep"
   where
      d = sdim tt
      joinC s = compose1 (finishH jt) $ app1 js s where (jt,js) = join (VList tt)
finishH t = toStr t
compose1 a b = "(" ++ a ++ "." ++ b ++ ")"
app1 a b = "(" ++ a ++ b ++ ")"
toStr VInt = inspect VInt
toStr VChr = "(:[])"
toStr (VList [VChr]) = " id "
toStr a = error $ "(internal) can't to str from " ++ show a
joinH t 0 = (vstr, "(const"++toStr t++")")
joinH (VList [e]) 1 = (vstr, "(\\a b->intercalate a $ map "++toStr e++" b)")
joinH (VList [VList e]) depth = (VList [rt], "(map."++ej++")") where (rt, ej)=joinH (VList e) (depth - 1)
joinH (VList (ts@(_:_))) 2 = (VList [vstr], "(\\a b->map (intercalate a . "++tuple2List (length ts) ++ "."++(appTuple $ map toStr ts) ++ ") b)")
joinH (VList (ts@(_:_))) depth = (VList rts, "(\\a b->map ("++appTuple rfsa++") b)") where
   (rts, rfs)=unzip $ map (\t->joinH t (depth-2)) ts
   rfsa = map (\f->f++" a ") rfs
joinH ts _ = (ts, "(\\_ b->b)")
join t = joinH t (sdim $ elemT t)
tuple2List n = tupleLambda n $ \args -> "["++intercalate "," args++"]"
vectorize op rtf [t1, VList t2] =
   (VList rts, "(\\a->map "++appTuple (map (\rop->"("++rop++"a)") rops)++")") where
      (rts, rops) = unzip $ map (\t->vectorize op rtf [t1,t]) t2
vectorize op rtf [t1, t2] = (rtf [t1, t2], "("++op++")")
fullVectorize lhsDim rhsDim = let
   extra = max lhsDim rhsDim
   lhsRepeat = cycleN "repeat" "a" (extra - lhsDim)
   rhsRepeat = cycleN "repeat" "b" (extra - rhsDim)
   cycleN s init n = " " ++ (if n < 0 then init
      else iterate (\inner->"(" ++ s ++ inner ++ ")") (" "++init) !! n)
   in
      ("(\\op a b->"++cycleN "zipWith" "op" extra ++ lhsRepeat ++ rhsRepeat ++")", extra)
isBaseElemChr VChr = True
isBaseElemChr (VList [e]) = isBaseElemChr e
isBaseElemChr _ = False
coerce2 [AutoType] t = t
coerce2 t [AutoType] = t
coerce2 [VChr] [VChr] = [VChr]
coerce2 [VInt] [VInt] = [VInt]
coerce2 [a] [b] | isNum a && isNum b = [vstr]
coerce2 [a] [VList [VChr]]
   | VChr == a || (not $ isBaseElemChr a) = [vstr]
   | otherwise = [a]
coerce2 [VList [VChr]] [a] = coerce2 [a] [vstr]
coerce2 [VList a] [b] | isNum b = [VList (coerce2 a [b])]
coerce2 [b] [VList a] | isNum b = coerce2 [VList a] [b]
coerce2 [VList a] [VList b] = [VList $ coerce2 a b]
coerce2 [a] (b:_) = coerce2 [a] [b]
coerce2 (a:_) [b] = coerce2 [a] [b]
coerce2 (a:as) (b:bs) = coerce2 [a] [b] ++ coerce2 as bs
coerceToH (a, AutoType) = "(const "++defaultValue [a]++")"
coerceToH (a, b) | a==b || (isNum a && isNum b) = "(id)"
coerceToH (VList [VInt], VList [VChr]) = "(id)"
coerceToH (VList [VChr], VInt) = "(sToA.show)"
coerceToH (VInt, VList [VChr]) = "(fromMaybe 0.readMaybe.aToS)"
coerceToH (VChr, VList a) = "(head."++coerceTo [VChr] a++")"
coerceToH (a, VList b) | csdim [a] < csdim [VList b] = "(concatMap"++coerceTo [a] b++")"
coerceToH (VList a, b) | baseElem (head a) == VInt && cidim [VList a] > cidim [b] = "((:[])."++coerceTo a [b]++")"
coerceToH (VList a, b) | baseElem (head a) == VChr && csdim [VList a] > csdim [b] = "((:[])."++coerceTo a [b]++")"
   = "(map"++coerceTo a b++")"
coerceTo to from = tupleLambda (length from) $ \args -> toTuple $
   zipWith3 (\t f a->"("++coerceToH(t,f)++" "++a++")") to from args ++ defaults
   where defaults = map defaultValue1 $ drop (length from) to
defaultValue1 (VInt) = "0"
defaultValue1 (VChr) = "space"
defaultValue1 (VList _) = "[]"
defaultValue = toTuple . map defaultValue1
baseElem (VList e) = baseElem $ head e
baseElem t = t
cidim [VInt] = 0
cidim [VChr] = 0
cidim [VList a] = 1+cidim a
cidim (t:_:_) = cidim [t]
sdim [VInt] = 1
sdim [VChr] = 0
sdim [VList a] = 1+sdim a
sdim (t:t2:ts) = 1+ maximum (map (sdim.(:[])) (t:t2:ts))
csdim [VInt] = 1
csdim [VChr] = 0
csdim [VList a] = 1+csdim a
csdim (t:_:_) = csdim [t]
appTuple ops = tupleLambda (length ops) $ \varNames -> (toTuple $ zipWith (++) ops varNames)
coerce leftType rightType =
   let
      coercedType = coerce2 leftType rightType
      coerceFn = coerceTo coercedType
   in (coercedType, coerceFn leftType, coerceFn rightType)
coerceEither leftType rightType =
   let
      coercedType = coerce2 leftType rightType
      coerceFn = coerceTo coercedType
   in
      (coercedType, "(either "++coerceFn leftType++coerceFn rightType++")")
promoteList [VList t] = (VList t, " id ")
promoteList t = (VList t, "(:[])")
promoteListRepeat [VList t] = (VList t, " id ")
promoteListRepeat t = (VList t, " repeat ")
flatten (VList [a]) = "concat." ++ flatten a
flatten _ = "(:[])"
fillAccums c n = "(\\f "++parenNonNothing (intercalate","varNames)++"->f "++parenNonNothing(intercalate","rhs)++")"
   where
      varNames = varNamesN (n-c)
      rhs = varNames ++ replicate c "()"
      parenNonNothing "" = ""
      parenNonNothing a = "("++a++")"
lazyOr a1 a2 = (a1, "\\a b->if "++truthy a1++" a then a else "++coerceTo a1 a2++"b")
lazyAnd a1 a2 = (a2, "\\a b->if "++truthy a1++" a then b else "++coerceTo a2 a1++"a")
   where
data SmartList a = SmartList Int (DList.DList a) deriving (Show, Eq, Ord)
newSmartList t = SmartList (length t) (DList.fromList t)
smartLength (SmartList len _) = len
smartAppend (SmartList l1 v1) (SmartList l2 v2) = SmartList (l1+l2) (DList.append v1 v2)
data VT = VInt | VChr | VList [VT] | VFn [VT] [VT] | AutoType | InvalidType
   deriving (Show, Eq)
vstr = VList [VChr]
isNum VInt = True
isNum VChr = True
isNum _ = False
isList (VList _) = True
isList _ = False
ret (VFn from to) = to
elemT (VList e) = e
elemT s = error $ "is not a list: " ++ show s
toTuple [t] = t
toTuple s = "(" ++ intercalate "," s ++ ")"
toHsType VInt = Just "Integer"
toHsType VChr = toHsType VInt
toHsType (VList ts) =
   toHsTypes ts >>= \s -> Just $ "["++s++"]"
toHsType (VFn a b) = Nothing
toHsType (ItWasAConstant) = Nothing
toHsType (AutoType) = Nothing
toHsType (InvalidType) = Nothing
toHsType (StaticInt _) = Nothing
toHsType e = error $ "cant toHsType " ++ show e
toHsTypes ts =
   let elems = map toHsType ts in
   if all isJust elems then
      Just $ toTuple $ catMaybes elems
   else Nothing
toHsReadType VInt = "Integer"
toHsReadType VChr = "Char"
toHsReadType (VList [VChr]) = "String"
toHsReadType (VList ts) = "["++toHsReadTypes ts++"]"
toHsReadType (VFn from to) = toHsReadTypes from ++ " -> " ++ toHsReadTypes to
toHsReadType e = error $ "cant toHsReadType " ++ show e
toHsReadTypes ts = toTuple $ map toHsReadType ts
xorChr [VInt, VChr] = VChr
xorChr [VChr, VInt] = VChr
xorChr _ = VInt
orChr [_, VChr] = VChr
orChr [VChr, _] = VChr
orChr _ = VInt
usage = "\
\Usage: nibbles [-c|-e|-v|-hs] [-simple] [filename] [args]*\n\
\\n\
\Nibbles - a functional code golf language for mortals.\n\
\\n\
\Documentation at "++webpage++"\n\
\\n\
\Modes:\n\
\  default = compile the nibbles program to out.hs and run it\n\
\  -c = compactify into bytes (2 nibbles each), save to base filename.nbb\n\
\  -e = expand and print the literate form\n\
\  -hs = only generate out.hs\n\
\  -v = version\n\
\\n\
\Other options:\n\
\  -simple = disable extensions and implicit things\n\
\\n\
\Filename and extension specifies source location and input format.\n\
\  .nbl file = literate form\n\
\  .nbb file = binary form\n\
\  empty = use stdin in literate form\n"
data ParseMode = FromLit | FromBytes deriving Eq
headerRaw = [litFile|Header.hs|]
main=do
   args <- getArgs
   let (progArgs, nonProgArgs) = partition isNibblesArg args
   let (opts,fileArgs) = partition isOpt nonProgArgs
   (contents, basename, parseMode) <- case fileArgs of
         [] -> do
            when (not$null progArgs) (hPutStrLn stderr "Warning: args are being treated as input args and nibbles code is being read from stdin")
            contents <- getContents
            return (contents, "a", FromLit)
         [f] -> case ext of
            ".nbb" -> do
               contents <- readFile f
               return (contents, base, FromBytes)
            ".nbl" -> do
               contents <- readFile f
               return (contents, base, FromLit)
            _ -> errorWithoutStackTrace "file extension must be .nbb or .nbl"
            where (base, ext) = splitExtension f
         e -> errorWithoutStackTrace $ "too many filename args:" ++ (show e) ++ "\n" ++ usage
   let ?isSimple = isSimple args
   let (cargs, reader) = toLetArgs progArgs
   let compileFn = compile finish "" cargs
   let (impl, bRaw, lit, litWarnings)  = compileFn $ case parseMode of
         FromLit -> Lit contents contents 0
         FromBytes -> Nib (concatMap fromByte contents) 0
   let paddedNibs = padToEvenNibbles bRaw
   let nibBytes = toBytes paddedNibs
   let (_,_,binLit,_) = compileFn (Nib (paddedNibs) 0)
   case opts of
      ops | null $ filter (\opt -> not (isOtherOption opt) && opt /= "-hs") ops -> do
         errored <- litErrorHandle "Warning" bRaw litWarnings
         maybeNibBytes <- if errored then return []
         else do
            when (parseMode == FromLit) $ do
               checkWouldExtractCorrectly binLit lit
               hPutStrLn stderr $ reverse $ reverse {-make strict to print at same time-} $
                  "size = " ++ (show $ length bRaw) ++ " nibbles ("++ (show $ div (length bRaw + 1) 2) ++ " bytes)"
            return nibBytes
         fullHs <- toFullHs impl maybeNibBytes reader
         writeFile "out.hs" fullHs
         when (ops /= ["-hs"]) $ runHs "out.hs" progArgs
      ["-c"] -> do
         checkWouldExtractCorrectly binLit lit
         errored <- litErrorHandle "Error" bRaw litWarnings
         when errored $ errorWithoutStackTrace "aborting"
         let outname = (basename ++ ".nbb")
         hPutStrLn stderr $ "wrote " ++ (show $ length nibBytes) ++ " bytes to " ++ outname
         writeFile outname nibBytes
      ["-e"] -> putStrLn lit
      ["-v"] -> putStrLn version
      e -> errorWithoutStackTrace $ "invalid option " ++ (show e) ++ "\n" ++ usage
estimateBinLength [] = 0
estimateBinLength (16:_:s) = 2 + estimateBinLength s
estimateBinLength (_:s) = 1 + estimateBinLength s
litErrorHandle msgPrefix bRaw litWarnings = do
   let binOnlyOps = any (==16) bRaw
   if binOnlyOps && null litWarnings then do
      hPutStrLn stderr $ msgPrefix ++ ": you are using literal only ops, estimated size: " ++ show (estimateBinLength bRaw) ++ " nibbles"
      return True
   else if not (null litWarnings) then do
      mapM_ (\msg -> hPutStrLn stderr $ msgPrefix++": " ++msg) litWarnings
      return True
   else return False
checkWouldExtractCorrectly binLit lit = do
   when (binLit /= lit) $ do
      hPutStrLn stderr "Warning: your code's binary would actual extract to:"
      hPutStrLn stderr binLit
      hPutStrLn stderr "instead of:"
      hPutStrLn stderr lit
      hPutStrLn stderr "Please report this as it is a bug (there is supposed to be a more clear accidental extension detection, this is just a fail safe)"
isOpt arg = isPrefixOf "-" arg
toBytes = map toByte . chunksOf 2
isOtherOption = flip elem ["-simple"]
isSimple = elem "-simple"
toFullHs impl nibBytes reader = do
   return $ header ++ "\n\
   \progSource = "++show nibBytes++"\n\
   \main=do\n\
   \ hSetEncoding stdin char8\n\
   \ hSetEncoding stdout char8\n\
   \ args <- getArgs \n\
   \ "++reader++"\n\
   \ interact ((\\input->let output=" ++ flatHs (implCode impl) ++ "\n\
   \  in aToS $ if output == sToA progSource\n\
   \    then output else finishLn output).sToA)"
runHs filename args = do
   setLocaleEncoding initLocaleEncoding
   (_, _, _, p) <- createProcess (proc "ghc" ["-O", filename]){ std_out = CreatePipe }
   ex <- waitForProcess p
   case ex of
      ExitSuccess -> callProcess "./out" args
      ExitFailure _ -> error "failed to compile hs (likely an internal nibbles bug, please report it!)"
